########################################################################
# File :   Optimizer.py
# Author : Stuart Paterson
########################################################################
"""
  The Optimizer base class is an agent that polls for jobs with a specific
  status and minor status pair.  The checkJob method is overridden for all
  optimizer instances and associated actions are performed there.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, exit as dExit

from DIRAC.AccountingSystem.Client.Types.Job import Job as AccountingJob
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client import JobMinorStatus
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.WorkloadManagementSystem.DB.JobLoggingDB import JobLoggingDB


class OptimizerModule(AgentModule):
    """
    The specific agents must provide the following methods:

      *  initialize() for initial settings
      *  beginExecution()
      *  execute() - the main method called in the agent cycle
      *  endExecution()
      *  finalize() - the graceful exit of the method, this one is usually used
               for the agent restart
    """

    #############################################################################
    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)
        self.jobDB = None
        self.elasticJobParametersDB = None
        self.logDB = None
        self.startingMinorStatus = None
        self.startingMajorStatus = JobStatus.CHECKING
        self.failedStatus = None
        self.requiredJobInfo = "jdl"
        self._initResult = None

    def initialize(self, jobDB=None, logDB=None):
        """Initialization of the Optimizer Agent."""
        self.jobDB = JobDB() if jobDB is None else jobDB
        if not self.jobDB.isValid():
            dExit(1)

        useESForJobParametersFlag = Operations().getValue("/Services/JobMonitoring/useESForJobParametersFlag", False)
        if useESForJobParametersFlag:
            try:
                result = ObjectLoader().loadObject(
                    "WorkloadManagementSystem.DB.ElasticJobParametersDB", "ElasticJobParametersDB"
                )
                if not result["OK"]:
                    return result
                self.elasticJobParametersDB = result["Value"]()
            except RuntimeError as excp:
                return S_ERROR("Can't connect to DB: %s" % excp)

        self.logDB = JobLoggingDB() if logDB is None else logDB

        optimizerName = self.am_getModuleParam("agentName")
        if optimizerName.endswith("Agent"):
            optimizerName = optimizerName[: -len("Agent")]
        self.am_setModuleParam("optimizerName", optimizerName)

        self.startingMinorStatus = self.am_getModuleParam("optimizerName")
        self.failedStatus = self.am_getOption("FailedJobStatus", JobStatus.FAILED)
        self.am_setOption("PollingTime", 30)

        return self.initializeOptimizer()

    def initializeOptimizer(self):
        """To be overwritten by inheriting class"""
        return S_OK()

    #############################################################################
    def execute(self):
        """The main agent execution method"""

        result = self.initializeOptimizer()
        if not result["OK"]:
            return result
        self._initResult = result["Value"]

        condition = {"Status": self.startingMajorStatus}
        if self.startingMinorStatus:
            condition["MinorStatus"] = self.startingMinorStatus

        result = self.jobDB.selectJobs(condition)
        if not result["OK"]:
            self.log.warn("Failed to get a job list from the JobDB")
            return S_ERROR("Failed to get a job list from the JobDB")

        if not result["Value"]:
            self.log.verbose("No pending jobs to process")
            return S_OK("No work to do")

        for job in result["Value"]:
            result = self.getJobDefinition(job)
            if not result["OK"]:
                self.setFailedJob(job, result["Message"], "")
                continue
            jobDef = result["Value"]
            result = self.optimizeJob(job, jobDef["classad"])

        return S_OK()

    #############################################################################
    def optimizeJob(self, job, classAdJob):
        """Call the corresponding Optimizer checkJob method"""
        self.log.verbose("Job will be processed", "%s by %sExecutor" % (job, self.am_getModuleParam("optimizerName")))
        result = self.checkJob(job, classAdJob)
        if not result["OK"]:
            self.setFailedJob(job, result["Message"], classAdJob)
        return result

    #############################################################################
    def getJobDefinition(self, job, jobDef=False):
        """Retrieve JDL of the Job and return jobDef dictionary"""
        if not jobDef:
            jobDef = {}
        # If not jdl in jobinfo load it
        if "jdl" not in jobDef:
            if self.requiredJobInfo == "jdlOriginal":
                result = self.jobDB.getJobJDL(job, original=True)
                if not result["OK"]:
                    self.log.error("No JDL for job", "%s" % job)
                    return S_ERROR("No JDL for job")
                jobDef["jdl"] = result["Value"]
            if self.requiredJobInfo == "jdl":
                result = self.jobDB.getJobJDL(job)
                if not result["OK"]:
                    self.log.error("No JDL for job", "%s" % job)
                    return S_ERROR("No JDL for job")
                jobDef["jdl"] = result["Value"]
        # Load the classad if needed
        if "jdl" in jobDef and "classad" not in jobDef:
            try:
                classad = ClassAd(jobDef["jdl"])
            except Exception:
                self.log.debug("Cannot load JDL")
                return S_ERROR(JobMinorStatus.ILLEGAL_JOB_JDL)
            if not classad.isOK():
                self.log.debug("Warning: illegal JDL for job %s, will be marked problematic" % (job))
                return S_ERROR(JobMinorStatus.ILLEGAL_JOB_JDL)
            jobDef["classad"] = classad
        return S_OK(jobDef)

    #############################################################################
    def getOptimizerJobInfo(self, job, reportName):
        """This method gets job optimizer information that will
        be used for
        """
        self.log.debug("self.jobDB.getJobOptParameter(%s,'%s')" % (job, reportName))
        result = self.jobDB.getJobOptParameter(job, reportName)
        if result["OK"]:
            value = result["Value"]
            if not value:
                self.log.warn("JobDB returned null value", "for %s %s" % (job, reportName))
                return S_ERROR("No optimizer info returned")
            else:
                try:
                    return S_OK(eval(value))
                except Exception as x:
                    return S_ERROR("Could not evaluate optimizer parameters: %s" % repr(x))

        return result

    #############################################################################
    def setOptimizerJobInfo(self, job, reportName, value):
        """This method sets the job optimizer information that will subsequently
        be used for job scheduling and TURL queries on the WN.
        """
        self.log.debug("self.jobDB.setJobOptParameter(%s,'%s','%s')" % (job, reportName, value))
        if not self.am_Enabled():
            return S_OK()
        return self.jobDB.setJobOptParameter(job, reportName, str(value))

    #############################################################################
    def setOptimizerChain(self, job, value):
        """This method sets the job optimizer chain, in principle only needed by
        one of the optimizers.
        """
        self.log.debug("self.jobDB.setOptimizerChain(%s,%s)" % (job, value))
        if not self.am_Enabled():
            return S_OK()
        return self.jobDB.setOptimizerChain(job, value)

    #############################################################################
    def setNextOptimizer(self, job):
        """This method is executed when the optimizer instance has successfully
        processed the job.  The next optimizer in the chain will subsequently
        start to work on the job.
        """

        result = self.logDB.addLoggingRecord(
            job,
            status=self.startingMajorStatus,
            minorStatus=self.startingMinorStatus,
            source=self.am_getModuleParam("optimizerName"),
        )
        if not result["OK"]:
            self.log.warn("Can't add logging record", result["Message"])

        self.log.debug("self.jobDB.setNextOptimizer(%s,'%s')" % (job, self.am_getModuleParam("optimizerName")))
        return self.jobDB.setNextOptimizer(job, self.am_getModuleParam("optimizerName"))

    #############################################################################
    def updateJobStatus(self, job, status, minorStatus=None, appStatus=None):
        """This method updates the job status in the JobDB, this should only be
        used to fail jobs due to the optimizer chain.
        """
        self.log.debug(
            "self.jobDB.setJobStatus(%s,'Status/Minor/Application','%s'/'%s'/'%s',update=True)"
            % (job, status, str(minorStatus), str(appStatus))
        )
        if not self.am_Enabled():
            return S_OK()

        result = self.jobDB.setJobStatus(job, status, minorStatus, appStatus)
        if not result["OK"]:
            return result

        result = self.logDB.addLoggingRecord(
            job,
            status=status,
            minorStatus=minorStatus,
            applicationStatus=appStatus,
            source=self.am_getModuleParam("optimizerName"),
        )
        if not result["OK"]:
            self.log.warn("Can't add logging record", result["Message"])

        return S_OK()

    #############################################################################
    def setJobParam(self, job, reportName, value):
        """This method updates a job parameter in the JobDB."""
        if not self.am_Enabled():
            return S_OK()

        self.log.debug("setJobParameter(%s,'%s','%s')" % (job, reportName, value))
        if self.elasticJobParametersDB:
            return self.elasticJobParametersDB.setJobParameter(job, reportName, value)
        return self.jobDB.setJobParameter(job, reportName, value)

    #############################################################################
    def setFailedJob(self, job, msg, classAdJob=None):
        """This method moves the job to the failed status"""
        self.log.debug("self.updateJobStatus(%s,'%s','%s')" % (job, self.failedStatus, msg))
        if not self.am_Enabled():
            return S_OK()
        self.updateJobStatus(job, self.failedStatus, msg)
        if classAdJob:
            self.sendAccountingRecord(job, msg, classAdJob)

    #############################################################################
    def checkJob(self, job, classad):
        """This method controls the checking of the job, should be overridden in a subclass"""
        self.log.warn("Optimizer: checkJob method should be implemented in a subclass")
        return S_ERROR("Optimizer: checkJob method should be implemented in a subclass")

    #############################################################################
    def sendAccountingRecord(self, job, msg, classAdJob):
        """
        Send and accounting record for the failed job
        """
        accountingReport = AccountingJob()
        accountingReport.setStartTime()
        accountingReport.setEndTime()

        owner = classAdJob.getAttributeString("Owner")
        userGroup = classAdJob.getAttributeString("OwnerGroup")
        jobGroup = classAdJob.getAttributeString("JobGroup")
        jobType = classAdJob.getAttributeString("JobType")
        jobClass = "unknown"
        if classAdJob.lookupAttribute("JobSplitType"):
            jobClass = classAdJob.getAttributeString("JobSplitType")
        inputData = []
        processingType = "unknown"
        if classAdJob.lookupAttribute("ProcessingType"):
            processingType = classAdJob.getAttributeString("ProcessingType")
        if classAdJob.lookupAttribute("InputData"):
            inputData = classAdJob.getListFromExpression("InputData")
        inputDataFiles = len(inputData)
        outputData = []
        if classAdJob.lookupAttribute("OutputData"):
            outputData = classAdJob.getListFromExpression("OutputData")
        outputDataFiles = len(outputData)

        acData = {
            "User": owner,
            "UserGroup": userGroup,
            "JobGroup": jobGroup,
            "JobType": jobType,
            "JobClass": jobClass,
            "ProcessingType": processingType,
            "FinalMajorStatus": self.failedStatus,
            "FinalMinorStatus": msg,
            "CPUTime": 0.0,
            "NormCPUTime": 0.0,
            "ExecTime": 0.0,
            "InputDataSize": 0.0,
            "OutputDataSize": 0.0,
            "InputDataFiles": inputDataFiles,
            "OutputDataFiles": outputDataFiles,
            "DiskSpace": 0.0,
            "InputSandBoxSize": 0.0,
            "OutputSandBoxSize": 0.0,
            "ProcessedEvents": 0.0,
        }

        self.log.debug("Accounting Report is:")
        self.log.debug(acData)
        accountingReport.setValuesFromDict(acData)
        return accountingReport.commit()



""" The HTCondor TimeLeft utility interrogates the HTCondor batch system for the
    current CPU consumed, as well as its limit.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os

from DIRAC import S_OK, S_ERROR
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import runCommand
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.ResourceUsage import ResourceUsage

__RCSID__ = "$Id$"


class HTCondorResourceUsage(ResourceUsage):
    """
    This is the HTCondor plugin of the TimeLeft Utility.
    HTCondor does not provide any way to get the wallclock/cpu limit, the batch system just provides fair-sharing to
    users and groups: the limit depends on many parameters.
    However, some Sites have introduced a MaxRuntime variable that sets a wallclock time limit to the allocations and
    allow us to get an estimation of the resources usage.
    """

    def __init__(self):
        """Standard constructor"""
        super(HTCondorResourceUsage, self).__init__("HTCondor", "_CONDOR_JOB_AD")

    def getResourceUsage(self):
        """Returns S_OK with a dictionary containing the entries WallClock, WallClockLimit, and Unit for current slot."""
        # $_CONDOR_JOB_AD corresponds to the path to the .job.ad file
        # It contains info about the job:
        # - MaxRuntime: wallclock time allocated to the job - not officially supported by HTCondor,
        #   only present on some Sites
        # - CurrentTime: current time
        # - JobCurrentStartDate: start of the job execution
        jobDescription = os.environ.get("_CONDOR_JOB_AD")
        cmd = "condor_status -ads %s -af MaxRuntime CurrentTime-JobCurrentStartDate" % jobDescription
        result = runCommand(cmd)
        if not result["OK"]:
            return S_ERROR("Current batch system is not supported")

        output = str(result["Value"]).split(" ")
        if len(output) != 2:
            self.log.warn("Cannot open $_CONDOR_JOB_AD: output probably empty")
            return S_ERROR("Current batch system is not supported")

        wallClockLimit = output[0]
        wallClock = output[1]
        if wallClockLimit == "undefined":
            self.log.warn("MaxRuntime attribute is not supported")
            return S_ERROR("Current batch system is not supported")

        wallClockLimit = float(wallClockLimit)
        wallClock = float(wallClock)

        consumed = {"WallClock": wallClock, "WallClockLimit": wallClockLimit, "Unit": "WallClock"}

        self.log.debug("TimeLeft counters complete:", str(consumed))
        return S_OK(consumed)



""" Test Encoding function of DIRAC
It contains tests for DISET and JSON.
Some tests can be passed by both, while some can only be passed by one.

Typically, we know JSON cannot serialize tuples, or integers as dictionary keys.
On the other hand, it can serialize some objects, while DISET cannot.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from string import printable
import datetime
import sys

from DIRAC.Core.Utilities.DEncode import encode as disetEncode, decode as disetDecode, g_dEncodeFunctions
from DIRAC.Core.Utilities.JEncode import encode as jsonEncode, decode as jsonDecode, JSerializable
from DIRAC.Core.Utilities.MixedEncode import encode as mixEncode, decode as mixDecode

from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import (
    builds,
    integers,
    lists,
    recursive,
    floats,
    text,
    booleans,
    none,
    dictionaries,
    tuples,
    dates,
    datetimes,
)
import six
from pytest import mark, approx, raises, fixture, skip

parametrize = mark.parametrize

# function_scoped_fixture is only used in Python 3 compatible release of hypothesis
if sys.version_info.major >= 3:
    function_scoped = (HealthCheck.function_scoped_fixture,)
else:
    function_scoped = tuple()

# List of couple (encoding, decoding) functions
# In order to test a new library, import the encode/decode
# function, and add the tuple here

disetTuple = (disetEncode, disetDecode)
jsonTuple = (jsonEncode, jsonDecode)
mixTuple = (mixEncode, mixDecode)

enc_dec_imp = (disetTuple, jsonTuple, (mixTuple, "No", "No"), (mixTuple, "Yes", "No"), (mixTuple, "Yes", "Yes"))
enc_dec_ids = (
    "disetTuple",
    "jsonTuple",
    "mixTuple",
    "mixTuple (DIRAC_USE_JSON_DECODE=Yes)",
    "mixTuple (DIRAC_USE_JSON_ENCODE=Yes",
)

enc_dec_imp_without_json = (disetTuple, (mixTuple, "No", "No"), (mixTuple, "Yes", "No"))
enc_dec_ids_without_json = ("disetTuple", "mixTuple", "mixTuple (DIRAC_USE_JSON_DECODE=Yes)")


def myDates():
    """We define a custom date strategy in order
    to pull date after 1900 (limitation of strftime)
    and without microseconds
    """
    return dates(
        min_value=datetime.datetime(1900, 1, 1, 0, 0).date(),
        max_value=datetime.datetime.max.date(),
    )


def myDatetimes():
    """We define a custom datetime strategy in order
    to pull date after 1900 (limitation of strftime)
    and without microseconds
    """
    # Build a strategy by removing the microsecond from a datetimes strategy
    # https://hypothesis.readthedocs.io/en/latest/data.html#hypothesis.strategies.builds
    return builds(
        lambda inDt: inDt.replace(microsecond=0),
        datetimes(min_value=datetime.datetime(1900, 1, 1, 0, 0), max_value=datetime.datetime.max, timezones=none()),
    )


# These initial strategies are the basic types supported by the original dEncode
# Unfortuately we cannot make nested structure with floats because as the floats
# are not stable, the result is approximative, and it becomes extremely difficult
# to compare
# Datetime also starts only at 1900 because earlier date can't be dumped with strftime
initialStrategies = none() | booleans() | text() | integers() | myDatetimes() | myDates()
initialJsonStrategies = none() | booleans() | text() | myDatetimes() | myDates()


# From a strategy (x), make a new strategy
# We basically use that to make nested structures
# see http://hypothesis.readthedocs.io/en/latest/data.html#recursive-data

nestedStrategy = recursive(initialStrategies, lambda x: lists(x) | dictionaries(text(), x) | tuples(x))

# This strategy does not return tuples
nestedStrategyJson = recursive(initialJsonStrategies, lambda x: lists(x) | dictionaries(text(), x))


def test_everyBaseTypeIsTested():
    """Make sure that each supported base type in the original
    DEncode module are tested here.
    We rely on the fact that the test function will be called
    "test_BaseType"
    """
    for encodeFunc in g_dEncodeFunctions.values():
        testFuncName = ("test_BaseType_%s" % encodeFunc.__name__).replace("encode", "")
        globals()[testFuncName]


def agnosticTestFunction(enc_dec_tuple, data):
    """Function called by all the other to test that
    decode(encode) returns the original data

    :param enc_dec: tuple of function (encoding, decoding)
    :param data: data to be worked on
    """
    encode, decode = enc_dec_tuple
    encodedData = encode(data)
    decodedData, lenData = decode(encodedData)

    assert data == decodedData
    assert lenData == len(encodedData)

    return decodedData


def base_enc_dec(request, monkeypatch):
    """base function to generate the (encoding, decoding) tuple and potentially setting the environment variables

    :param request: fixture request. Will contain either an encoding/decoding tuple,
     or the same tuple plus two env variables to be set
    """

    if len(request.param) == 2:
        return request.param

    enc_dec_tuple, use_json_decode, use_json_encode = request.param
    monkeypatch.setenv("DIRAC_USE_JSON_DECODE", use_json_decode)
    monkeypatch.setenv("DIRAC_USE_JSON_ENCODE", use_json_encode)
    return enc_dec_tuple


@fixture(scope="function", params=enc_dec_imp, ids=enc_dec_ids)
def enc_dec(request, monkeypatch):
    """Fixture to generate the (encoding, decoding) tuple for all the serialization types"""
    return base_enc_dec(request, monkeypatch)


@fixture(scope="function", params=enc_dec_imp_without_json, ids=enc_dec_ids_without_json)
def enc_dec_without_json(request, monkeypatch):
    """Fixture to generate the (encoding, decoding) tuple for all the serialization types which
    do not involve json encoding
    """
    return base_enc_dec(request, monkeypatch)


@settings(suppress_health_check=function_scoped)
@given(data=booleans())
def test_BaseType_Bool(enc_dec, data):
    """Test for boolean"""
    agnosticTestFunction(enc_dec, data)


@settings(suppress_health_check=function_scoped)
@given(data=myDates())
def test_BaseType_Dates(enc_dec, data):
    """Test for date"""
    agnosticTestFunction(enc_dec, data)


@settings(suppress_health_check=function_scoped)
@given(data=myDatetimes())
def test_BaseType_DateTime(enc_dec, data):
    """Test for datetime"""
    agnosticTestFunction(enc_dec, data)


# Json does not serialize keys as integers but as string
@settings(suppress_health_check=function_scoped)
@given(data=dictionaries(integers(), integers()))
def test_BaseType_Dict(enc_dec_without_json, data):
    """Test for basic dict"""
    agnosticTestFunction(enc_dec_without_json, data)


@settings(suppress_health_check=function_scoped)
@given(data=integers(max_value=sys.maxsize))
def test_BaseType_Int(enc_dec, data):
    """Test for integer"""
    agnosticTestFunction(enc_dec, data)


# CAUTION: DEncode is not precise for floats !!
@settings(suppress_health_check=function_scoped)
@given(data=floats(allow_nan=False))
def test_BaseType_Float(enc_dec, data):
    """Test that float is approximatly stable"""
    encode, decode = enc_dec
    encodedData = encode(data)
    decodedData, lenData = decode(encodedData)
    assert data == approx(decodedData)
    assert lenData == len(encodedData)


@settings(suppress_health_check=function_scoped)
@given(data=lists(integers()))
def test_BaseType_List(enc_dec, data):
    """Test for List"""
    agnosticTestFunction(enc_dec, data)


@settings(suppress_health_check=function_scoped)
@given(data=integers(min_value=sys.maxsize + 1))
def test_BaseType_Long(enc_dec, data):
    """Test long type"""
    agnosticTestFunction(enc_dec, data)


def test_BaseType_None(
    enc_dec,
):
    """Test None case"""
    agnosticTestFunction(enc_dec, None)


@settings(suppress_health_check=function_scoped)
@given(data=text(printable))
def test_BaseType_String(enc_dec, data):
    """Test basic strings"""
    # we need to cast to str because text() returns unicode
    data = str(data)
    agnosticTestFunction(enc_dec, data)


# Tuple are not serialized in JSON
@settings(suppress_health_check=function_scoped)
@given(data=tuples(integers()))
def test_BaseType_Tuple(enc_dec_without_json, data):
    """Test basic tuple"""
    agnosticTestFunction(enc_dec_without_json, data)


@settings(suppress_health_check=function_scoped)
@given(data=text())
def test_BaseType_Unicode(enc_dec, data):
    """Test unicode data"""
    agnosticTestFunction(enc_dec, data)


# Json will not pass this because of tuples and integers as dict keys
@mark.slow
@settings(suppress_health_check=function_scoped)
@given(data=nestedStrategy)
def test_nestedStructure(enc_dec_without_json, data):
    """Test nested structure"""
    agnosticTestFunction(enc_dec_without_json, data)


# DEncode raises KeyError.....
# Others raise TypeError
# @parametrize('enc_dec', enc_dec_imp)
def test_NonSerializable(enc_dec):
    """Test that a class that does not inherit from the serializable class
    raises TypeError
    """

    class NonSerializable(object):
        """Dummy class not serializable"""

        pass

    data = NonSerializable()
    with raises((TypeError, KeyError)):
        agnosticTestFunction(enc_dec, data)


class Serializable(JSerializable):
    """Dummy class inheriting from JSerializable"""

    _attrToSerialize = ["instAttr"]

    def __init__(self, instAttr=None):
        self.instAttr = instAttr

    def __eq__(self, other):
        return all([getattr(self, attr) == getattr(other, attr) for attr in self._attrToSerialize])


@mark.slow
@settings(suppress_health_check=(HealthCheck.too_slow,) + function_scoped)
@given(data=nestedStrategyJson)
def test_Serializable(data):
    """Test if a simple serializable class with one random argument
    can be serialized
    """

    objData = Serializable(instAttr=data)

    agnosticTestFunction(jsonTuple, objData)


def test_nonDeclaredAttr():
    """Tests that an argument not in the list of arguments to serialized
    is not serialized
    """

    objData = Serializable()
    objData.notToBeSerialized = 1

    encodedData = jsonEncode(objData)
    decodedData, _lenData = jsonDecode(encodedData)

    assert not hasattr(decodedData, "notToBeSerialized")


class BadSerializable(JSerializable):
    """Missing _attrToSerialize attribute"""

    pass


def test_missingAttrToSerialize():
    """Tests that an argument not in the list of arguments to serialized
    is not serialized
    """

    objData = BadSerializable()

    with raises(TypeError):
        agnosticTestFunction(jsonTuple, objData)


@mark.slow
@settings(suppress_health_check=function_scoped)
@given(data=nestedStrategyJson)
def test_nestedSerializable(data):
    """Test that a serializable containing a serializable class
    can be serialized
    """

    subObj = Serializable(instAttr=data)
    objData = Serializable(instAttr=subObj)
    agnosticTestFunction(jsonTuple, objData)


def test_types():
    """Ensure that the DEncode types object matches the Python 2 types module"""
    if not six.PY2:
        skip("This test only makes sense on Python 2")
    import types as pythonTypes
    from DIRAC.Core.Utilities.DEncode import types as DIRACTypes

    assert DIRACTypes.IntType is pythonTypes.IntType  # pylint: disable=no-member
    assert DIRACTypes.LongType is pythonTypes.LongType  # pylint: disable=no-member
    assert DIRACTypes.FloatType is pythonTypes.FloatType  # pylint: disable=no-member
    assert DIRACTypes.BooleanType is pythonTypes.BooleanType  # pylint: disable=no-member
    assert DIRACTypes.StringType is pythonTypes.StringType  # pylint: disable=no-member
    assert DIRACTypes.UnicodeType is pythonTypes.UnicodeType  # pylint: disable=no-member
    assert DIRACTypes.NoneType is pythonTypes.NoneType  # pylint: disable=no-member
    assert DIRACTypes.ListType is pythonTypes.ListType  # pylint: disable=no-member
    assert DIRACTypes.TupleType is pythonTypes.TupleType  # pylint: disable=no-member
    assert DIRACTypes.DictType is pythonTypes.DictType  # pylint: disable=no-member



#!/usr/bin/env python
"""
Get metadata for the given file specified by its Logical File Name or for a list of files
contained in the specifed file

Example:
  $ dirac-dms-catalog-metadata /formation/user/v/vhamar/Example.txt
  FileName                                     Size        GUID                                     Status   Checksum
  /formation/user/v/vhamar/Example.txt         34          EDE6DDA4-3344-3F39-A993-8349BA41EB23     1        eed20d47
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC import exit as DIRACExit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Names"))
    Script.registerArgument(["Catalog:   file catalog plug-ins"], mandatory=False)
    Script.parseCommandLine()

    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

    import os

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    inputFileName, catalogs = Script.getPositionalArgs(group=True)

    if os.path.exists(inputFileName):
        inputFile = open(inputFileName, "r")
        string = inputFile.read()
        lfns = string.splitlines()
        inputFile.close()
    else:
        lfns = [inputFileName]

    res = FileCatalog(catalogs=catalogs).getFileMetadata(lfns)
    if not res["OK"]:
        print("ERROR:", res["Message"])
        DIRACExit(-1)

    print("FileName".ljust(100), "Size".ljust(10), "GUID".ljust(40), "Status".ljust(8), "Checksum".ljust(10))
    for lfn in sorted(res["Value"]["Successful"].keys()):
        metadata = res["Value"]["Successful"][lfn]
        checksum = ""
        if "Checksum" in metadata:
            checksum = str(metadata["Checksum"])
        size = ""
        if "Size" in metadata:
            size = str(metadata["Size"])
        guid = ""
        if "GUID" in metadata:
            guid = str(metadata["GUID"])
        status = ""
        if "Status" in metadata:
            status = str(metadata["Status"])
        print("%s %s %s %s %s" % (lfn.ljust(100), size.ljust(10), guid.ljust(40), status.ljust(8), checksum.ljust(10)))

    for lfn in sorted(res["Value"]["Failed"].keys()):
        message = res["Value"]["Failed"][lfn]
        print(lfn, message)


if __name__ == "__main__":
    main()



""" Some utilities for FTS3...
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import random
import threading

from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations as opHelper
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus


def _checkSourceReplicas(ftsFiles, preferDisk=True):
    """Check the active replicas

    :params ftsFiles: list of FT3Files
    :param preferDisk: (default True) prefer disk replicas if available
                        (see :py:meth:`DIRAC.DataManagementSystem.Client.DataManager.DataManager.getActiveReplicas`)

    :returns: Successful/Failed {lfn : { SE1 : PFN1, SE2 : PFN2 } , ... }
    """

    lfns = list(set([f.lfn for f in ftsFiles]))
    res = DataManager().getActiveReplicas(lfns, getUrl=False, preferDisk=preferDisk)

    return res


def selectUniqueSource(ftsFiles, fts3Plugin, allowedSources=None):
    """
    For a list of FTS3files object, select a random source, and group the files by source.

    We also return the FTS3Files for which we had problems getting replicas

    :param ftsFiles: list of FTS3File object
    :param fts3Plugin: plugin instance to use to chose between sources
    :param allowedSources: list of allowed sources


    :return:  S_OK(({ sourceSE: [ FTS3Files] }, {FTS3File: errors}))

    """

    _log = gLogger.getSubLogger("selectUniqueSource")

    # destGroup will contain for each target SE a dict { source : [list of FTS3Files] }
    groupBySource = {}

    # For all files, check which possible sources they have
    # If we specify allowedSources, don't restrict the choice to disk replicas
    res = _checkSourceReplicas(ftsFiles, preferDisk=not allowedSources)
    if not res["OK"]:
        return res

    filteredReplicas = res["Value"]

    # LFNs for which we failed to get replicas
    failedFiles = {}

    for ftsFile in ftsFiles:

        # If we failed to get the replicas, add the FTS3File to
        # the dictionnary
        if ftsFile.lfn in filteredReplicas["Failed"]:
            errMsg = filteredReplicas["Failed"][ftsFile.lfn]
            failedFiles[ftsFile] = errMsg
            _log.debug("Failed to get active replicas", "%s,%s" % (ftsFile.lfn, errMsg))
            continue

        replicaDict = filteredReplicas["Successful"][ftsFile.lfn]

        try:
            uniqueSource = fts3Plugin.selectSourceSE(ftsFile, replicaDict, allowedSources)
            groupBySource.setdefault(uniqueSource, []).append(ftsFile)
        except ValueError as e:
            _log.info("No allowed replica source for file", "%s: %s" % (ftsFile.lfn, repr(e)))
            continue

    return S_OK((groupBySource, failedFiles))


def groupFilesByTarget(ftsFiles):
    """
    For a list of FTS3files object, group the Files by target

    :param ftsFiles: list of FTS3File object
    :return: {targetSE : [ ftsFiles] } }

    """

    # destGroup will contain for each target SE a dict { possible source : transfer metadata }
    destGroup = {}

    for ftsFile in ftsFiles:
        destGroup.setdefault(ftsFile.targetSE, []).append(ftsFile)

    return S_OK(destGroup)


def getFTS3Plugin(vo=None):
    """
    Return an instance of the FTS3Plugin configured in the CS

    :param vo: vo config to look for
    """
    pluginName = opHelper(vo=vo).getValue("DataManagement/FTSPlacement/FTS3/FTS3Plugin", "Default")

    objLoader = ObjectLoader()
    _class = objLoader.loadObject(
        "DataManagementSystem.private.FTS3Plugins.%sFTS3Plugin" % pluginName, "%sFTS3Plugin" % pluginName
    )

    if not _class["OK"]:
        raise Exception(_class["Message"])

    fts3Plugin = _class["Value"](vo=vo)
    return fts3Plugin


threadLocal = threading.local()


class FTS3ServerPolicy(object):
    """
    This class manages the policy for choosing a server
    """

    def __init__(self, serverDict, serverPolicy="Random"):
        """
        Call the init of the parent, and initialize the list of FTS3 servers
        """

        self.log = gLogger.getSubLogger("FTS3ServerPolicy")

        self._serverDict = serverDict
        self._serverList = list(serverDict)
        self._maxAttempts = len(self._serverList)
        self._nextServerID = 0
        self._resourceStatus = ResourceStatus()

        methName = "_%sServerPolicy" % serverPolicy.lower()
        if not hasattr(self, methName):
            self.log.error("Unknown server policy %s. Using Random instead" % serverPolicy)
            methName = "_randomServerPolicy"

        self._policyMethod = getattr(self, methName)

    def _failoverServerPolicy(self, _attempt):
        """
        Returns always the server at a given position (normally the first one)

        :param attempt: position of the server in the list
        """
        if _attempt >= len(self._serverList):
            raise Exception("FTS3ServerPolicy.__failoverServerPolicy: attempt to reach non existing server index")
        return self._serverList[_attempt]

    def _sequenceServerPolicy(self, _attempt):
        """
        Every time the this policy is called, return the next server on the list
        """

        fts3server = self._serverList[self._nextServerID]
        self._nextServerID = (self._nextServerID + 1) % len(self._serverList)
        return fts3server

    def _randomServerPolicy(self, _attempt):
        """
        return a server from shuffledServerList
        """

        if getattr(threadLocal, "shuffledServerList", None) is None:
            threadLocal.shuffledServerList = self._serverList[:]
            random.shuffle(threadLocal.shuffledServerList)

        fts3Server = threadLocal.shuffledServerList[_attempt]

        if _attempt == self._maxAttempts - 1:
            random.shuffle(threadLocal.shuffledServerList)

        return fts3Server

    def _getFTSServerStatus(self, ftsServer):
        """Fetch the status of the FTS server from RSS"""

        res = self._resourceStatus.getElementStatus(ftsServer, "FTS")
        if not res["OK"]:
            return res

        result = res["Value"]
        if ftsServer not in result:
            return S_ERROR("No FTS Server %s known to RSS" % ftsServer)

        if result[ftsServer]["all"] == "Active":
            return S_OK(True)

        return S_OK(False)

    def chooseFTS3Server(self):
        """
        Choose the appropriate FTS3 server depending on the policy
        """

        fts3Server = None
        attempt = 0

        while not fts3Server and attempt < self._maxAttempts:

            fts3Server = self._policyMethod(attempt)
            res = self._getFTSServerStatus(fts3Server)

            if not res["OK"]:
                self.log.warn("Error getting the RSS status for %s: %s" % (fts3Server, res))
                fts3Server = None
                attempt += 1
                continue

            ftsServerStatus = res["Value"]

            if not ftsServerStatus:
                self.log.warn("FTS server %s is not in good shape. Choose another one" % fts3Server)
                fts3Server = None
                attempt += 1

        if fts3Server:
            return S_OK(self._serverDict[fts3Server])

        return S_ERROR("Could not find an FTS3 server (max attempt reached)")



"""
Create and put Requests to move files.

List of operations:

#. ReplicateAndRegister LFNs
#. Check for Migration
#. Remove all other replicas for these files
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation

sLog = gLogger.getSubLogger("CreateMoving")
__RCSID__ = "$Id$"


class CreateMovingRequest(object):
    """Create the request to move files from one SE to another."""

    def __init__(self):
        """Constructor."""
        self.requests = []
        self._reqClient = None
        self._fcClient = None
        self._requestName = "Moving_"
        self.switches = {}
        self.options = [
            ("L", "List", "File containing list of LFNs to move"),
            ("P", "Path", "LFN path to folder, all files in the folder will be moved"),
            ("S", "SourceSE", "Where to remove the LFNs from"),
            ("T", "TargetSE", "Where to move the LFNs to"),
            ("N", "Name", "Name of the Request"),
        ]
        self.flags = [
            ("C", "CheckMigration", "Ensure the LFNs are migrated to tape before removing any replicas"),
            ("X", "Execute", "Put Requests, else dryrun"),
        ]
        self.registerSwitchesAndParseCommandLine()
        self.getLFNList()
        self.getLFNMetadata()

    @property
    def fcClient(self):
        """Return FileCatalogClient."""
        if not self._fcClient:
            from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

            self._fcClient = FileCatalog()
        return self._fcClient

    @property
    def reqClient(self):
        """Return RequestClient."""
        if not self._reqClient:
            from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient

            self._reqClient = ReqClient()
        return self._reqClient

    @property
    def dryRun(self):
        """Return dry run flag."""
        return self.switches["DryRun"]

    @property
    def targetSE(self):
        """Return the list of targetSE."""
        return self.switches["TargetSE"]

    @property
    def sourceSEs(self):
        """Return the list of sourceSEs."""
        return self.switches["SourceSE"]

    @property
    def lfnFolderPath(self):
        """Return the lfn folder path where to find the files of the request."""
        return self.switches.get("Path", None)

    def registerSwitchesAndParseCommandLine(self):
        """Register the default plus additional parameters and parse options.

        :param list options: list of three tuple for options to add to the script
        :param list flags:  list of three tuple for flags to add to the script
        :param str opName
        """
        for short, longOption, doc in self.options:
            Script.registerSwitch(short + ":" if short else "", longOption + "=", doc)
        for short, longOption, doc in self.flags:
            Script.registerSwitch(short, longOption, doc)
            self.switches[longOption] = False
        Script.parseCommandLine()
        if Script.getPositionalArgs():
            Script.showHelp(exitCode=1)

        for switch in Script.getUnprocessedSwitches():
            for short, longOption, doc in self.options:
                if switch[0] == short or switch[0].lower() == longOption.lower():
                    sLog.verbose("Found switch %r with value %r" % (longOption, switch[1]))
                    self.switches[longOption] = switch[1]
                    break
            for short, longOption, doc in self.flags:
                if switch[0] == short or switch[0].lower() == longOption.lower():
                    self.switches[longOption] = True
                    break

        self.checkSwitches()
        self.switches["DryRun"] = not self.switches.get("Execute", False)
        self.switches["SourceSE"] = self.switches.get("SourceSE", "").split(",")

    def checkSwitches(self):
        """Check the switches, set autoName if needed."""
        if not self.switches.get("SourceSE"):
            raise RuntimeError('Have to set "SourceSE"')
        if not self.switches.get("TargetSE"):
            raise RuntimeError('Have to set "TargetSE"')
        if not self.switches.get("List") and not self.switches.get("Path"):
            raise RuntimeError('Have to set "List" or "Path"')

    def getLFNList(self):
        """Get list of LFNs.

        Either read the provided file, or get the files found beneath the provided folder.

        :returns: list of lfns
        :raises: RuntimeError, ValueError
        """
        if self.switches.get("List"):
            if os.path.exists(self.switches.get("List")):
                self.lfnList = list(
                    set([line.split()[0] for line in open(self.switches.get("List")).read().splitlines()])
                )
            else:
                raise ValueError("%s not a file" % self.switches.get("List"))
        elif self.lfnFolderPath:
            path = self.lfnFolderPath
            sLog.debug("Check if %r is a directory" % path)
            isDir = returnSingleResult(self.fcClient.isDirectory(path))
            sLog.debug("Result: %r" % isDir)
            if not isDir["OK"] or not isDir["Value"]:
                sLog.error("Path is not a directory", isDir.get("Message", ""))
                raise RuntimeError("Path %r is not a directory" % path)
            sLog.notice("Looking for files in %r" % path)

            metaDict = {"SE": self.sourceSEs[0]} if self.switches.get("SourceOnly") else {}
            lfns = self.fcClient.findFilesByMetadata(metaDict=metaDict, path=path)
            if not lfns["OK"]:
                sLog.error("Could not find files")
                raise RuntimeError(lfns["Message"])
            self.lfnList = lfns["Value"]

        if self.lfnList:
            sLog.notice("Will create request(s) with %d lfns" % len(self.lfnList))
            if len(self.lfnList) == 1:
                raise RuntimeError("Only 1 file in the list, aborting!")
            return

        raise ValueError('"Path" or "List" need to be provided!')

    def getLFNMetadata(self):
        """Get the metadata for all the LFNs."""
        metaData = self.fcClient.getFileMetadata(self.lfnList)
        error = False
        if not metaData["OK"]:
            sLog.error("Unable to read metadata for lfns: %s" % metaData["Message"])
            raise RuntimeError("Could not read metadata: %s" % metaData["Message"])

        self.metaData = metaData["Value"]
        for failedLFN, reason in self.metaData["Failed"].items():
            sLog.error("skipping %s: %s" % (failedLFN, reason))
            error = True
        if error:
            raise RuntimeError("Could not read all metadata")

        for lfn in self.metaData["Successful"].keys():
            sLog.verbose("found %s" % lfn)

    def run(self):
        """Perform checks and create the request."""
        from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator

        for count, lfnChunk in enumerate(breakListIntoChunks(self.lfnList, 100)):
            if not lfnChunk:
                sLog.error("LFN list is empty!!!")
                return 1

            requestName = "%s_%d" % (self.switches.get("Name"), count)
            request = self.createRequest(requestName, lfnChunk)
            valid = RequestValidator().validate(request)
            if not valid["OK"]:
                sLog.error("putRequest: request not valid", "%s" % valid["Message"])
                return 1
            else:
                self.requests.append(request)

        self.putOrRunRequests()
        return 0

    def createRequest(self, requestName, lfnChunk):
        """Create the Request."""
        request = Request()
        request.RequestName = requestName

        replicate = Operation()
        replicate.Type = "ReplicateAndRegister"
        replicate.TargetSE = self.switches.get("TargetSE")
        self.addLFNs(replicate, lfnChunk, addPFN=True)
        request.addOperation(replicate)

        if self.switches.get("CheckMigration"):
            checkMigration = Operation()
            checkMigration.Type = "CheckMigration"
            checkMigration.TargetSE = self.switches.get("TargetSE")
            self.addLFNs(checkMigration, lfnChunk, addPFN=True)
            request.addOperation(checkMigration)

        removeReplicas = Operation()
        removeReplicas.Type = "RemoveReplica"
        removeReplicas.TargetSE = ",".join(self.switches.get("SourceSE", []))
        self.addLFNs(removeReplicas, lfnChunk)
        request.addOperation(removeReplicas)

        return request

    def addLFNs(self, operation, lfns, addPFN=False):
        """Add lfns to operation.

        :param operation: the operation instance to which the files will be added
        :param list lfns: list of lfns
        :param bool addPFN: if true adds PFN to each File
        """
        if not self.metaData:
            self.getLFNMetadata()

        for lfn in lfns:
            metaDict = self.metaData["Successful"][lfn]
            opFile = File()
            opFile.LFN = lfn
            if addPFN:
                opFile.PFN = lfn
            opFile.Size = metaDict["Size"]
            if "Checksum" in metaDict:
                # should check checksum type, now assuming Adler32 (metaDict['ChecksumType'] = 'AD')
                opFile.Checksum = metaDict["Checksum"]
                opFile.ChecksumType = "ADLER32"
            operation.addFile(opFile)

    def putOrRunRequests(self):
        """Run or put requests."""
        requestIDs = []

        if self.dryRun:
            sLog.notice("Would have created %d requests" % len(self.requests))
            for reqID, req in enumerate(self.requests):
                sLog.notice("Request %d:" % reqID)
                for opID, op in enumerate(req):
                    sLog.notice("        Operation %d: %s #lfn %d" % (opID, op.Type, len(op)))
            return 0
        for request in self.requests:
            putRequest = self.reqClient.putRequest(request)
            if not putRequest["OK"]:
                sLog.error("unable to put request %r: %s" % (request.RequestName, putRequest["Message"]))
                continue
            requestIDs.append(str(putRequest["Value"]))
            sLog.always("Request %r has been put to ReqDB for execution." % request.RequestName)

        if requestIDs:
            sLog.always("%d requests have been put to ReqDB for execution" % len(requestIDs))
            sLog.always("RequestID(s): %s" % " ".join(requestIDs))
            sLog.always("You can monitor the request status using the command: dirac-rms-request <requestName/ID>")
            return 0

        sLog.error("No requests created")
        return 1


@Script()
def main():
    try:
        CMR = CreateMovingRequest()
        CMR.run()
    except Exception as e:
        if LogLevels.getLevelValue(sLog.getLevel()) <= LogLevels.VERBOSE:
            sLog.exception("Failed to create Moving Request")
        else:
            sLog.error("ERROR: Failed to create Moving Request:", str(e))
        exit(1)
    exit(0)


if __name__ == "__main__":
    main()



"""
  Service interface adapted to work with tornado, must be used only by tornado service handlers
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from tornado import gen
from tornado.ioloop import IOLoop
from DIRAC.ConfigurationSystem.private.ServiceInterfaceBase import ServiceInterfaceBase
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC import gLogger


class ServiceInterfaceTornado(ServiceInterfaceBase):
    """
    Service interface adapted to work with tornado
    """

    def __init__(self, sURL):
        ServiceInterfaceBase.__init__(self, sURL)

    def _launchCheckSlaves(self):
        """
        Start loop to check if slaves are alive
        """
        IOLoop.current().spawn_callback(self.run)
        gLogger.info("Starting purge slaves thread")

    def run(self):
        """
        Check if slaves are alive
        """
        while True:
            yield gen.sleep(gConfigurationData.getSlavesGraceTime())
            self._checkSlavesStatus()

    def _updateServiceConfiguration(self, urlSet, fromMaster=False):
        """
        Update configuration in a set of service in parallel

        :param set urlSet: a set of service URLs
        :param fromMaster: flag to force updating from the master CS
        :return: Nothing
        """

        for url in urlSet:
            IOLoop.current().spawn_callback(self._forceServiceUpdate, [url, fromMaster])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

###########################################################
#
#  Simple executor script for Batch class methods.
#
#  The script is concatenated on the fly with the required
#  batch system class definition.
#
#  NB: This scipt is executed using the local (to the WN)
#  python version, so support for py2 and py3 is necessary.
#
#  15.11.2014
#  Author: A.T.
#
###########################################################

executeBatchContent = """
if __name__ == "__main__":

  import sys
  import json
  import traceback
  try:
    from six.moves.urllib.parse import quote as urlquote
    from six.moves.urllib.parse import unquote as urlunquote
  except ImportError:
    try:
      from urllib import unquote as urlunquote
      from urllib import quote as urlquote
    except ImportError:
      from urllib.parse import quote as urlquote
      from urllib.parse import unquote as urlunquote


  arguments = sys.argv[1]
  inputDict = json.loads(urlunquote(arguments))

  method = inputDict.pop('Method')
  batchSystem = inputDict.pop('BatchSystem')
  batch = locals()[batchSystem]()

  try:
    result = getattr(batch, method)(**inputDict)
  except Exception:
    result = traceback.format_exc()

  resultJson = urlquote(json.dumps(result))
  print("============= Start output ===============")
  print(resultJson)
"""



"""Helpers for working with extensions to DIRAC"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import argparse
from collections import defaultdict
import fnmatch
import functools
import glob
import importlib
import os
import pkgutil
import sys

import importlib_resources
import six
import DIRAC

try:
    from importlib.machinery import PathFinder
except ImportError:
    # Fallback for Python 2
    import imp

    class ModuleSpec(object):
        def __init__(self, name, submodule_search_locations):
            self.name = name
            self.submodule_search_locations = submodule_search_locations

    class PathFinder(object):
        @classmethod
        def find_spec(cls, name, path=None):
            try:
                _, pathname, _ = imp.find_module(name, path)
            except ImportError:
                return None
            else:
                return ModuleSpec(name, [pathname])


def iterateThenSort(func):
    """Convenience decorator the find* functions

    Converts a function that takes a python module into one which accepts one or
    modules/module names. The returned values are converted to a sorted list of
    unique values.
    """

    @functools.wraps(func)
    def newFunc(modules, *args, **kwargs):
        if not isinstance(modules, (list, tuple)):
            modules = [modules]

        results = set()
        for module in modules:
            if isinstance(module, six.string_types):
                module = importlib.import_module(module)
            results |= set(func(module, *args, **kwargs))
        return sorted(results)

    return newFunc


@iterateThenSort
def findSystems(module):
    """Find the systems for one or more DIRAC extension(s)

    :param list/str/module module: One or more Python modules or Python module names
    :returns: list of system names
    """
    return {x.name for x in _findSystems(module)}


def findAgents(modules):
    """Find the agents for one or more DIRAC extension(s)

    :param list/str/module module: One or more Python modules or Python module names
    :returns: list of tuples of the form (SystemName, AgentName)
    """
    return findModules(modules, "Agent", "*Agent")


def findExecutors(modules):
    """Find the executors for one or more DIRAC extension(s)

    :param list/str/module module: One or more Python modules or Python module names
    :returns: list of tuples of the form (SystemName, ExecutorName)
    """
    return findModules(modules, "Executor")


def findServices(modules):
    """Find the services for one or more DIRAC extension(s)

    :param list/str/module module: One or more Python modules or Python module names
    :returns: list of tuples of the form (SystemName, ServiceName)
    """
    return findModules(modules, "Service", "*Handler")


@iterateThenSort
def findDatabases(module):
    """Find the DB SQL schema defintions for one or more DIRAC extension(s)

    :param list/str/module module: One or more Python modules or Python module names
    :returns: list of tuples of the form (SystemName, dbSchemaFilename)
    """
    # This can be "fn.name" when DIRAC is Python 3 only
    return {(system.name, os.path.basename(str(fn))) for system, fn in _findFile(module, "DB", "*DB.sql")}


@iterateThenSort
def findModules(module, submoduleName, pattern="*"):
    """Find the direct submodules from one or more DIRAC extension(s) that match a pattern

    :param list/str/module module: One or more Python modules or Python module names
    :param str submoduleName: The submodule under ``module`` in which to look
    :param str pattern: A ``fnmatch``-style pattern that the submodule must match
    :returns: list of tuples of the form (SystemName, ServiceName)
    """
    for system in _findSystems(module):
        agentModule = PathFinder.find_spec(submoduleName, path=system.submodule_search_locations)
        if not agentModule:
            continue
        for _, name, _ in pkgutil.iter_modules(agentModule.submodule_search_locations):
            if fnmatch.fnmatch(name, pattern):
                yield system.name, name


def entrypointToExtension(entrypoint):
    """ "Get the extension name from an EntryPoint object"""
    # In Python 3.9 this can be "entrypoint.module"
    module = entrypoint.pattern.match(entrypoint.value).groupdict()["module"]
    return module.split(".")[0]


def extensionsByPriority():
    """Get the list of installed extensions"""
    if six.PY3:
        return _extensionsByPriorityPy3()
    else:
        return _extensionsByPriorityPy2()


def _extensionsByPriorityPy2():
    initList = glob.glob(os.path.join(DIRAC.rootPath, "*DIRAC", "__init__.py"))
    extensions = [os.path.basename(os.path.dirname(k)) for k in initList]
    # Return the extensions, sorting such that vanilla DIRAC is always last
    # It's not correct but it's less incorrect than ComponentInstaller.getExtensions
    return sorted(extensions, key=lambda x: (x == "DIRAC", x))


def _extensionsByPriorityPy3():
    """Discover extensions using the setuptools metadata"""
    # This is Python 3 only, Python 2 installations should never try to use this
    from importlib import metadata  # pylint: disable=no-name-in-module

    priorties = defaultdict(list)
    for entrypoint in set(metadata.entry_points()["dirac"]):
        extensionName = entrypointToExtension(entrypoint)
        extension_metadata = entrypoint.load()()
        priorties[extension_metadata["priority"]].append(extensionName)

    extensions = []
    for priority, extensionNames in sorted(priorties.items(), reverse=True):
        if len(extensionNames) != 1:
            print(
                "WARNING: Found multiple extensions with priority",
                "{} ({})".format(priority, extensionNames),
                file=sys.stderr,
            )
        # If multiple are passed, sort the extensions so things are deterministic at least
        extensions.extend(sorted(extensionNames))
    return extensions


def getExtensionMetadata(extensionName):
    """Get the metadata for a given extension name"""
    # This is Python 3 only, Python 2 installations should never try to use this
    from importlib import metadata  # pylint: disable=no-name-in-module

    for entrypoint in metadata.entry_points()["dirac"]:
        if extensionName == entrypointToExtension(entrypoint):
            return entrypoint.load()()


def recurseImport(modName, parentModule=None, hideExceptions=False):
    from DIRAC import S_OK, S_ERROR, gLogger

    if parentModule is not None:
        raise NotImplementedError(parentModule)
    try:
        return S_OK(importlib.import_module(modName))
    except ImportError as excp:
        if str(excp).startswith("No module named"):
            return S_OK()
        errMsg = "Can't load %s" % modName
        if not hideExceptions:
            gLogger.exception(errMsg)
        return S_ERROR(errMsg)


def _findSystems(module):
    """Implementation of _findSystems that returns a generator of system names"""
    for _, name, _ in pkgutil.iter_modules(module.__path__):
        if name.endswith("System"):
            yield PathFinder.find_spec(name, path=module.__path__)


def _findFile(module, submoduleName, pattern="*"):
    """Implementation of findDatabases"""
    for system in _findSystems(module):
        try:
            dbModule = importlib_resources.files(".".join([module.__name__, system.name, submoduleName]))
        except ImportError:
            continue
        for file in dbModule.iterdir():
            if fnmatch.fnmatch(file.name, pattern):
                yield system, file


def parseArgs():
    """CLI interface for use with the DIRAC integration tests"""
    parser = argparse.ArgumentParser()
    if six.PY3:
        subparsers = parser.add_subparsers(required=True, dest="function")
    else:
        subparsers = parser.add_subparsers()
    defaultExtensions = extensionsByPriority()
    for func in [findSystems, findAgents, findExecutors, findServices, findDatabases]:
        subparser = subparsers.add_parser(func.__name__)
        subparser.add_argument("--extensions", nargs="+", default=defaultExtensions)
        subparser.set_defaults(func=func)
    args = parser.parse_args()
    # Get the result and print it
    extensions = [importlib.import_module(e) for e in args.extensions]
    for result in args.func(extensions):
        if not isinstance(result, str):
            result = " ".join(result)
        print(result)


if __name__ == "__main__":
    parseArgs()



#!/usr/bin/env python
########################################################################
# File :    dirac-utils-file-adler
########################################################################
"""
Calculate alder32 of the supplied file

Example:
  $ dirac-utils-file-adler Example.tgz
  Example.tgz 88b4ca8b
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["File:     File Name"])
    _, files = Script.parseCommandLine(ignoreErrors=False)

    exitCode = 0

    import DIRAC
    from DIRAC.Core.Utilities.Adler import fileAdler

    for fa in files:
        adler = fileAdler(fa)
        if adler:
            print(fa.rjust(100), adler.ljust(10))  # pylint: disable=no-member
        else:
            print("ERROR %s: Failed to get adler" % fa)
            exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Legend encapsulates a graphical plot legend drawing tool

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from matplotlib.patches import Rectangle
from matplotlib.text import Text
from DIRAC.Core.Utilities.Graphs.GraphUtilities import *
from DIRAC.Core.Utilities.Graphs.Palette import Palette
from DIRAC.Core.Utilities.Graphs.GraphData import GraphData
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg

import six


class Legend(object):
    def __init__(self, data=None, axes=None, *aw, **kw):

        self.text_size = 0
        self.column_width = 0
        self.labels = {}
        if isinstance(data, dict):
            for label, ddict in data.items():
                # self.labels[label] = pretty_float(max([ float(x) for x in ddict.values() if x ]) )
                self.labels[label] = "%.1f" % max([float(x) for x in ddict.values() if x])
        elif isinstance(data, GraphData):
            self.labels = data.getLabels()
        else:
            self.labels = data
        # self.labels.reverse()
        self.ax = axes
        self.canvas = None
        if self.ax:
            self.canvas = self.ax.figure.canvas
            self.ax.set_axis_off()
        self.prefs = evalPrefs(*aw, **kw)
        self.palette = Palette()

        if self.labels and self.labels[0][0] != "NoLabels":
            percent_flag = self.prefs.get("legend_unit", "")
            if percent_flag == "%":
                sum_value = sum(data.label_values)
                if sum_value > 0.0:
                    self.labels = [(l, v / sum_value * 100.0) for l, v in self.labels]
        self.__get_column_width()

    def dumpPrefs(self):

        for key in self.prefs:
            print(key.rjust(20), ":", str(self.prefs[key]).ljust(40))

    def setLabels(self, labels):

        self.labels = labels

    def setAxes(self, axes):

        self.ax = axes
        self.canvas = self.ax.figure.canvas
        self.ax.set_axis_off()

    def getLegendSize(self):
        self.__get_column_width()
        legend_position = self.prefs["legend_position"]
        legend_width = float(self.prefs["legend_width"])
        legend_height = float(self.prefs["legend_height"])
        legend_padding = float(self.prefs["legend_padding"])
        legend_text_size = self.prefs.get("legend_text_size", self.prefs["text_size"])
        legend_text_padding = self.prefs.get("legend_text_padding", self.prefs["text_padding"])
        if legend_position in ["right", "left"]:
            # One column in case of vertical legend
            legend_width = self.column_width + legend_padding
            nLabels = len(self.labels)
            legend_max_height = nLabels * (legend_text_size + legend_text_padding)
        elif legend_position == "bottom":
            nColumns = min(self.prefs["legend_max_columns"], int(legend_width / self.column_width))
            nLabels = len(self.labels)
            maxRows = self.prefs["legend_max_rows"]
            nRows_ax = int(legend_height / 1.6 / self.prefs["text_size"])
            nRows_label = nLabels / nColumns + (nLabels % nColumns != 0)
            nRows = int(max(1, min(min(nRows_label, maxRows), nRows_ax)))
            text_padding = self.prefs["text_padding"]
            text_padding = pixelToPoint(text_padding, self.prefs["dpi"])
            legend_height = int(min(legend_height, (nRows * (self.text_size + text_padding) + text_padding)))
            legend_max_height = int(nLabels * (self.text_size + text_padding))
        return legend_width, legend_height, legend_max_height

    def __get_legend_text_size(self):

        text_size = self.prefs["text_size"]
        text_padding = self.prefs["text_padding"]
        legend_text_size = self.prefs.get("legend_text_size", text_size)
        legend_text_padding = self.prefs.get("legend_text_padding", text_padding)
        return legend_text_size, legend_text_padding

    def __get_column_width(self):

        max_length = 0
        max_column_text = ""
        flag = self.prefs.get("legend_numbers", True)
        unit = self.prefs.get("legend_unit", False)
        for label, num in self.labels:
            if not flag:
                num = None
            if num is not None:
                column_length = len(str(label) + str(num)) + 1
            else:
                column_length = len(str(label)) + 1
            if column_length > max_length:
                max_length = column_length
                if flag:
                    if isinstance(num, six.integer_types):
                        numString = str(num)
                    else:
                        numString = "%.1f" % float(num)
                    max_column_text = "%s  %s" % (str(label), numString)
                    if unit:
                        max_column_text += "%"
                else:
                    max_column_text = "%s   " % str(label)

        figure = Figure()
        canvas = FigureCanvasAgg(figure)
        dpi = self.prefs["dpi"]
        figure.set_dpi(dpi)
        l_size, _ = self.__get_legend_text_size()
        self.text_size = pixelToPoint(l_size, dpi)
        text = Text(0.0, 0.0, text=max_column_text, size=self.text_size)
        text.set_figure(figure)
        bbox = text.get_window_extent(canvas.get_renderer())
        columnwidth = bbox.width + 6 * l_size
        # make sure the legend fit in the box
        self.column_width = (
            columnwidth if columnwidth <= self.prefs["legend_width"] else self.prefs["legend_width"] - 6 * l_size
        )

    def draw(self):

        dpi = self.prefs["dpi"]
        ax_xsize = self.ax.get_window_extent().width
        ax_ysize = self.ax.get_window_extent().height
        nLabels = len(self.labels)
        nColumns = min(self.prefs["legend_max_columns"], int(ax_xsize / self.column_width))

        maxRows = self.prefs["legend_max_rows"]
        nRows_ax = int(ax_ysize / 1.6 / self.prefs["text_size"])
        nRows_label = nLabels / nColumns + (nLabels % nColumns != 0)
        nRows = max(1, min(min(nRows_label, maxRows), nRows_ax))
        self.ax.set_xlim(0.0, float(ax_xsize))
        self.ax.set_ylim(-float(ax_ysize), 0.0)

        legend_text_size, legend_text_padding = self.__get_legend_text_size()
        legend_text_size_point = pixelToPoint(legend_text_size, dpi)

        box_width = legend_text_size
        legend_offset = (ax_xsize - nColumns * self.column_width) / 2

        nc = 0
        # self.labels.reverse()

        for label, num in self.labels:
            num_flag = self.prefs.get("legend_numbers", True)
            percent_flag = self.prefs.get("legend_unit", "")
            if num_flag:
                if percent_flag == "%":
                    num = "%.1f" % num + "%"
                else:
                    num = "%.1f" % num
            else:
                num = None
            color = self.palette.getColor(label)
            row = nc % nRows
            column = int(nc / nRows)
            if row == nRows - 1 and column == nColumns - 1 and nc != nLabels - 1:
                last_text = "... plus %d more" % (nLabels - nc)
                self.ax.text(
                    float(column * self.column_width) + legend_offset,
                    -float(row * 1.6 * box_width),
                    last_text,
                    horizontalalignment="left",
                    verticalalignment="top",
                    size=legend_text_size_point,
                )
                break
            else:
                self.ax.text(
                    float(column * self.column_width) + 2.0 * box_width + legend_offset,
                    -row * 1.6 * box_width,
                    str(label),
                    horizontalalignment="left",
                    verticalalignment="top",
                    size=legend_text_size_point,
                )
                if num is not None:
                    self.ax.text(
                        float((column + 1) * self.column_width) - 2 * box_width + legend_offset,
                        -float(row * 1.6 * box_width),
                        str(num),
                        horizontalalignment="right",
                        verticalalignment="top",
                        size=legend_text_size_point,
                    )
                box = Rectangle(
                    (float(column * self.column_width) + legend_offset, -float(row * 1.6 * box_width) - box_width),
                    box_width,
                    box_width,
                )
                box.set_edgecolor("black")
                box.set_linewidth(pixelToPoint(0.5, dpi))
                box.set_facecolor(color)
                self.ax.add_patch(box)
                nc += 1



#!/usr/bin/env python
"""
Utilities for using M2Crypto SSL with DIRAC.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import tempfile
from M2Crypto import SSL, m2, X509

from DIRAC.Core.DISET import DEFAULT_SSL_CIPHERS, DEFAULT_SSL_METHODS
from DIRAC.Core.Security import Locations
from DIRAC.Core.Security.m2crypto.X509Chain import X509Chain

# Verify depth of peer certs
VERIFY_DEPTH = 50
DEBUG_M2CRYPTO = os.getenv("DIRAC_DEBUG_M2CRYPTO", "No").lower() in ("yes", "true")


def __loadM2SSLCTXHostcert(ctx):
    """Load hostcert & key from the default location and set them as the
    credentials for SSL context ctx.
    Returns None.
    """
    certKeyTuple = Locations.getHostCertificateAndKeyLocation()
    if not certKeyTuple:
        raise RuntimeError("Hostcert/key location not set")
    hostcert, hostkey = certKeyTuple
    if not os.path.isfile(hostcert):
        raise RuntimeError("Hostcert file (%s) is missing" % hostcert)
    if not os.path.isfile(hostkey):
        raise RuntimeError("Hostkey file (%s) is missing" % hostkey)
    # Make sure we never stall on a password prompt if the hostkey has a password
    # by specifying a blank string.
    ctx.load_cert(hostcert, hostkey, callback=lambda: "")


def __loadM2SSLCTXProxy(ctx, proxyPath=None):
    """Load proxy from proxyPath (or default location if not specified) and
    set it as the certificate & key to use for this SSL context.
    Returns None.
    """
    if not proxyPath:
        proxyPath = Locations.getProxyLocation()
    if not proxyPath:
        raise RuntimeError("Proxy location not set")
    if not os.path.isfile(proxyPath):
        raise RuntimeError("Proxy file (%s) is missing" % proxyPath)
    # See __loadM2SSLCTXHostcert for description of why lambda is needed.
    ctx.load_cert_chain(proxyPath, proxyPath, callback=lambda: "")


def ssl_verify_callback_print_error(ok, store):
    """This callback method does nothing but printing the error.
    It prints a few more useful info than the exception

    :param ok: current validation status
    :param store: pointer to the X509_CONTEXT_STORE
    """
    errnum = store.get_error()
    if errnum:
        print("SSL DEBUG ERRNUM %s ERRMSG %s" % (errnum, m2.x509_get_verify_error(errnum)))  # pylint: disable=no-member
    return ok


def getM2SSLContext(ctx=None, **kwargs):
    """Gets an M2Crypto.SSL.Context configured using the standard
    DIRAC connection keywords from kwargs. The keywords are:

      - clientMode: Boolean, if False hostcerts are always used. If True
                             a proxy is used unless other flags are set.
      - useCertificates: Boolean, Set to true to use hostcerts in client
                         mode.
      - proxyString: String, allow a literal proxy string to be provided.
      - proxyLocation: String, Path to file to use as proxy, defaults to
                               usual location(s) if not set.
      - skipCACheck: Boolean, if True, don't verify peer certificates.
      - sslMethods: String, List of SSL algorithms to enable in OpenSSL style
                            cipher format, e.g. "SSLv3:TLSv1".
      - sslCiphers: String, OpenSSL style cipher string of ciphers to allow
                            on this connection.

    If an existing context "ctx" is provided, it is just reconfigured with
    the selected arguments.

    Returns the new or updated context.
    """
    if not ctx:
        ctx = SSL.Context()
    # Set certificates for connection
    # CHRIS: I think clientMode was just an internal of pyGSI implementation
    # if kwargs.get('clientMode', False) and not kwargs.get('useCertificates', False):
    # if not kwargs.get('useCertificates', False):
    if kwargs.get("bServerMode", False) or kwargs.get("useCertificates", False):
        # Server mode always uses hostcert
        __loadM2SSLCTXHostcert(ctx)

    else:
        # Client mode has a choice of possible options
        if kwargs.get("proxyString", None):
            # M2Crypto cannot take an inmemory location or a string, so
            # so write it to a temp file and use proxyLocation
            with tempfile.NamedTemporaryFile(mode="wb") as tmpFile:
                tmpFilePath = tmpFile.name
                tmpFile.write(kwargs["proxyString"])
                __loadM2SSLCTXProxy(ctx, proxyPath=tmpFilePath)
        else:
            # Use normal proxy
            __loadM2SSLCTXProxy(ctx, proxyPath=kwargs.get("proxyLocation", None))

    verify_callback = ssl_verify_callback_print_error if DEBUG_M2CRYPTO else None

    # Set peer verification
    if kwargs.get("skipCACheck", False):
        # Don't validate peer, but still request creds
        ctx.set_verify(SSL.verify_none, VERIFY_DEPTH, callback=verify_callback)
    else:
        # Do validate peer
        ctx.set_verify(SSL.verify_peer | SSL.verify_fail_if_no_peer_cert, VERIFY_DEPTH, callback=verify_callback)
        # Set CA location
        caPath = Locations.getCAsLocation()
        if not caPath:
            raise RuntimeError("Failed to find CA location")
        if not os.path.isdir(caPath):
            raise RuntimeError("CA path (%s) is not a valid directory" % caPath)
        ctx.load_verify_locations(capath=caPath)

    # If the version of M2Crypto is recent enough, there is an API
    # to accept proxy certificate, and we do not need to rely on
    # OPENSSL_ALLOW_PROXY_CERT environment variable
    # which was removed as of openssl 1.1
    # We need this to be merged in M2Crypto: https://gitlab.com/m2crypto/m2crypto/merge_requests/236
    # We set the proper verify flag to the X509Store of the context
    # as described here https://www.openssl.org/docs/man1.1.1/man7/proxy-certificates.html
    if hasattr(SSL, "verify_allow_proxy_certs"):
        ctx.get_cert_store().set_flags(SSL.verify_allow_proxy_certs)  # pylint: disable=no-member
    # As of M2Crypto 0.37, the `verify_allow_proxy_certs` flag was moved
    # to X509 (https://gitlab.com/m2crypto/m2crypto/-/merge_requests/238)
    # It is more consistent with all the other flags,
    # but pySSL had it in SSL. Well...
    if hasattr(X509, "verify_allow_proxy_certs"):
        ctx.get_cert_store().set_flags(X509.verify_allow_proxy_certs)  # pylint: disable=no-member

    # Other parameters
    sslMethods = kwargs.get("sslMethods", DEFAULT_SSL_METHODS)
    if sslMethods:
        # Pylint can't see the m2 constants due to the way the library is loaded
        # We just have to disable that warning for the next bit...
        # pylint: disable=no-member
        methods = [("SSLv2", m2.SSL_OP_NO_SSLv2), ("SSLv3", m2.SSL_OP_NO_SSLv3), ("TLSv1", m2.SSL_OP_NO_TLSv1)]
        allowed_methods = sslMethods.split(":")
        # If a method isn't explicitly allowed, set the flag to disable it...
        for method, method_flag in methods:
            if method not in allowed_methods:
                ctx.set_options(method_flag)
        # SSL_OP_NO_SSLv2, SSL_OP_NO_SSLv3, SSL_OP_NO_TLSv1
    ciphers = kwargs.get("sslCiphers", DEFAULT_SSL_CIPHERS)
    ctx.set_cipher_list(ciphers)

    # log the debug messages
    if DEBUG_M2CRYPTO:
        ctx.set_info_callback()

    return ctx


def getM2PeerInfo(conn):
    """Gets the details of the current peer as a standard dict. The peer
    details are obtained from the supplied M2 SSL Connection obj "conn".
    The details returned are those from ~X509Chain.getCredentials, without Registry info:

       DN - Full peer DN as string
       x509Chain - Full chain of peer
       isProxy - Boolean, True if chain ends with proxy
       isLimitedProxy - Boolean, True if chain ends with limited proxy
       group - String, DIRAC group for this peer, if known

    Returns a dict of details.
    """
    chain = X509Chain.generateX509ChainFromSSLConnection(conn)
    creds = chain.getCredentials(withRegistryInfo=False)
    if not creds["OK"]:
        raise RuntimeError("Failed to get SSL peer info (%s)." % creds["Message"])
    peer = creds["Value"]

    peer["x509Chain"] = chain
    isProxy = chain.isProxy()
    if not isProxy["OK"]:
        raise RuntimeError("Failed to get SSL peer isProxy (%s)." % isProxy["Message"])
    peer["isProxy"] = isProxy["Value"]

    if peer["isProxy"]:
        peer["DN"] = creds["Value"]["identity"]
    else:
        peer["DN"] = creds["Value"]["subject"]

    isLimited = chain.isLimitedProxy()
    if not isLimited["OK"]:
        raise RuntimeError("Failed to get SSL peer isProxy (%s)." % isLimited["Message"])
    peer["isLimitedProxy"] = isLimited["Value"]

    return peer



""" Test class for FileCatalogComponents
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access

# imports
from mock import MagicMock

from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryManager.DirectoryTreeBase import DirectoryTreeBase
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryManager.DirectoryLevelTree import DirectoryLevelTree

# from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectorySimpleTree import DirectorySimpleTree
# from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryFlatTree import DirectoryFlatTree
# from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryNodeTree import DirectoryNodeTree

from DIRAC.DataManagementSystem.DB.FileCatalogComponents.FileManager.FileManagerBase import FileManagerBase

dbMock = MagicMock()
ugManagerMock = MagicMock()
ugManagerMock.getUserAndGroupID.return_value = {"OK": True, "Value": ("l_uid", "l_gid")}
dbMock.ugManager = ugManagerMock


####################################################################################
# TreeBase

dtb = DirectoryTreeBase()
dtb.db = dbMock


def test_Base_makeDirectory():
    res = dtb.makeDirectory("/path", {})
    assert res["OK"] is False  # this will need to be implemented on a derived class


####################################################################################
# LevelTree

dlt = DirectoryLevelTree()
dlt.db = dbMock


def test_Level_makeDirectory():
    res = dlt.makeDirectory("/path", {})
    assert res["OK"] is True  # this will need to be implemented on a derived class


####################################################################################
# SimpleTree
# FIXME: this fails... is it a genuine failure?

# dst = DirectorySimpleTree()
# dst.db = dbMock


# def test_Simple_makeDirectory():
#   res = dst.makeDirectory('/path', {})
#   assert res['OK'] is True  # this will need to be implemented on a derived class


####################################################################################
# FlatTree
# FIXME: this fails... is it a genuine failure?

# dft = DirectoryFlatTree()
# dft.db = dbMock


# def test_Flat_makeDirectory():
#   res = dft.makeDirectory('/path', {})
#   assert res['OK'] is True  # this will need to be implemented on a derived class


####################################################################################
# NodeTree
# FIXME: this fails... is it a genuine failure?

# dnt = DirectoryNodeTree()
# dnt.db = dbMock


# def test_Node_makeDirectory():
#   res = dnt.makeDirectory('/path', {})

#   assert res['OK'] is True  # this will need to be implemented on a derived class


####################################################################################
####################################################################################
# FileManagerBase

fmb = FileManagerBase()
fmb.db = dbMock


def test_Base_addFile():
    res = fmb.addFile({}, {})
    assert res["OK"] is True  # this will need to be implemented on a derived class, but it anyway returns S_OK()

    res = fmb.addFile({"aa": "aaa/bbb"}, {})
    assert res["OK"] is True  # this will need to be implemented on a derived class, but it anyway returns S_OK()
    assert "aa" in res["Value"]["Failed"]



""" This is the guy that you should use when you develop a script that interacts with DIRAC

    And don't forget to call parseCommandLine()
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys
import os.path
import inspect

from DIRAC.ConfigurationSystem.Client.LocalConfiguration import LocalConfiguration
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.Core.Utilities.DErrno import includeExtensionErrors

localCfg = LocalConfiguration()

caller = inspect.currentframe().f_back.f_globals["__name__"]


# There are several ways to call DIRAC scripts:
# * dirac-do-something
# * dirac-do-something.py
# * python dirac-do-something.py
# * pytest test-dirac-do-something.py
# * pytest --option test-dirac-do-something.py
# * pytest test-dirac-do-something.py::class::method
# The following lines attempt to keep only what is necessary to DIRAC, leaving the rest to pytest or whatever is before

i = 0
if caller != "__main__":
    # This is for the case when, for example, you run a DIRAC script from within pytest
    gLogger.debug("Called from module", caller)
    # Loop over until one of the argument is the caller, that is the DIRAC script
    for arg in sys.argv:
        # if the form is "pytest test-dirac-do-something.py::class::method", we
        # need to isolate the test-dirac-do-something.py
        arg = arg.split("::")[0]
        if os.path.basename(arg).replace(".py", "") == caller.split(".")[-1]:
            break
        i += 1

# If we reached the end, assume the caller is the first argument
i = 0 if i == len(sys.argv) else i
# Same thing here, get rid of the pytest specific class:meth options
scriptName = os.path.basename(sys.argv[i].split("::")[0]).replace(".py", "")
# The first argument DIRAC should parse is the next one
localCfg.firstOptionIndex = i + 1

gIsAlreadySetUsageMsg = False
gIsAlreadyInitialized = False


def parseCommandLine(script=False, ignoreErrors=False, initializeMonitor=False):
    global gIsAlreadySetUsageMsg, gIsAlreadyInitialized

    # Read and parse the script __doc__ to create a draft help message
    if not gIsAlreadySetUsageMsg:
        try:
            localCfg.setUsageMessage(inspect.currentframe().f_back.f_globals["__doc__"])
        except KeyError:
            pass
        gIsAlreadySetUsageMsg = True

    if gIsAlreadyInitialized:
        return False
    gLogger.showHeaders(False)

    return initialize(script, ignoreErrors, initializeMonitor, True)


def initialize(script=False, ignoreErrors=False, initializeMonitor=False, enableCommandLine=False):
    global scriptName, gIsAlreadyInitialized

    # Please do not call initialize in every file
    if gIsAlreadyInitialized:
        return False
    gIsAlreadyInitialized = True

    userDisabled = not localCfg.isCSEnabled()
    if not userDisabled:
        localCfg.disableCS()

    if not enableCommandLine:
        localCfg.disableParsingCommandLine()

    if script:
        scriptName = script
    localCfg.setConfigurationForScript(scriptName)

    if not ignoreErrors:
        localCfg.addMandatoryEntry("/DIRAC/Setup")
    resultDict = localCfg.loadUserData()
    if not ignoreErrors and not resultDict["OK"]:
        gLogger.error("There were errors when loading configuration", resultDict["Message"])
        sys.exit(1)

    if not userDisabled:
        localCfg.enableCS()

    if initializeMonitor:
        gMonitor.setComponentType(gMonitor.COMPONENT_SCRIPT)
        gMonitor.setComponentName(scriptName)
        gMonitor.setComponentLocation("script")
        gMonitor.initialize()
    else:
        gMonitor.disable()
    includeExtensionErrors()

    return True


def registerSwitch(showKey, longKey, helpString, callback=False):
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.registerCmdOpt`."""
    localCfg.registerCmdOpt(showKey, longKey, helpString, callback)


def registerArgument(description, mandatory=True, values=None, default=None):
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.registerCmdArg`."""
    localCfg.registerCmdArg(description, mandatory, values, default)


def getPositionalArgs(group=False):
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.getPositionalArguments`."""
    return localCfg.getPositionalArguments(group)


def getExtraCLICFGFiles():
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.getExtraCLICFGFiles`."""
    return localCfg.getExtraCLICFGFiles()


def getUnprocessedSwitches():
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.getUnprocessedSwitches`."""
    return localCfg.getUnprocessedSwitches()


def addDefaultOptionValue(option, value):
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.addDefaultEntry`."""
    localCfg.addDefaultEntry(option, value)


def setUsageMessage(usageMessage):
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.setUsageMessage`."""
    global gIsAlreadySetUsageMsg
    gIsAlreadySetUsageMsg = True
    try:
        localCfg.setUsageMessage(inspect.currentframe().f_back.f_globals["__doc__"])
    except KeyError:
        pass
    localCfg.setUsageMessage(usageMessage)


def disableCS():
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.disableCS`."""
    localCfg.disableCS()


def enableCS():
    """See :func:`~DIRAC.ConfigurationSystem.Client.LocalConfiguration.LocalConfiguration.enableCS`."""
    return localCfg.enableCS()


showHelp = localCfg.showHelp



""" CurveGraph represents simple line graphs with markers.

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.Graphs.PlotBase import PlotBase
from DIRAC.Core.Utilities.Graphs.GraphUtilities import (
    darkenColor,
    to_timestamp,
    PrettyDateLocator,
    PrettyDateFormatter,
    PrettyScalarFormatter,
)
from matplotlib.lines import Line2D
from matplotlib.dates import date2num
import datetime


class CurveGraph(PlotBase):

    """
    The CurveGraph class is a straightforward line graph with markers
    """

    def __init__(self, data, ax, prefs, *args, **kw):

        PlotBase.__init__(self, data, ax, prefs, *args, **kw)

    def draw(self):

        PlotBase.draw(self)
        self.x_formatter_cb(self.ax)

        if self.gdata.isEmpty():
            return None

        start_plot = 0
        end_plot = 0
        if "starttime" in self.prefs and "endtime" in self.prefs:
            start_plot = date2num(datetime.datetime.fromtimestamp(to_timestamp(self.prefs["starttime"])))
            end_plot = date2num(datetime.datetime.fromtimestamp(to_timestamp(self.prefs["endtime"])))

        labels = self.gdata.getLabels()
        labels.reverse()

        # If it is a simple plot, no labels are used
        # Evaluate the most appropriate color in this case
        if self.gdata.isSimplePlot():
            labels = [("SimplePlot", 0.0)]
            color = self.prefs.get("plot_color", "Default")
            if color.find("#") != -1:
                self.palette.setColor("SimplePlot", color)
            else:
                labels = [(color, 0.0)]

        tmp_max_y = []
        tmp_min_y = []
        tmp_x = []
        for label, num in labels:
            xdata = []
            ydata = []
            xerror = []
            yerror = []
            color = self.palette.getColor(label)
            plot_data = self.gdata.getPlotNumData(label)
            for key, value, error in plot_data:
                if value is None:
                    continue
                tmp_x.append(key)
                tmp_max_y.append(value + error)
                tmp_min_y.append(value - error)
                xdata.append(key)
                ydata.append(value)
                xerror.append(0.0)
                yerror.append(error)

            linestyle = self.prefs.get("linestyle", "-")
            marker = self.prefs.get("marker", "o")
            markersize = self.prefs.get("markersize", 8.0)
            markeredgewidth = self.prefs.get("markeredgewidth", 1.0)
            if not self.prefs.get("error_bars", False):
                line = Line2D(
                    xdata,
                    ydata,
                    color=color,
                    linewidth=1.0,
                    marker=marker,
                    linestyle=linestyle,
                    markersize=markersize,
                    markeredgewidth=markeredgewidth,
                    markeredgecolor=darkenColor(color),
                )
                self.ax.add_line(line)
            else:
                self.ax.errorbar(
                    xdata,
                    ydata,
                    color=color,
                    linewidth=2.0,
                    marker=marker,
                    linestyle=linestyle,
                    markersize=markersize,
                    markeredgewidth=markeredgewidth,
                    markeredgecolor=darkenColor(color),
                    xerr=xerror,
                    yerr=yerror,
                    ecolor=color,
                )

        ymax = max(tmp_max_y)
        ymax *= 1.1
        ymin = min(tmp_min_y, 0.0)
        ymin *= 1.1
        if "log_yaxis" in self.prefs:
            ymin = 0.001

        xmax = max(tmp_x) * 1.1
        if self.log_xaxis:
            xmin = 0.001
        else:
            xmin = 0

        ymin = self.prefs.get("ymin", ymin)
        ymax = self.prefs.get("ymax", ymax)
        xmin = self.prefs.get("xmin", xmin)
        xmax = self.prefs.get("xmax", xmax)

        self.ax.set_xlim(xmin=xmin, xmax=xmax)
        self.ax.set_ylim(ymin=ymin, ymax=ymax)
        if self.gdata.key_type == "time":
            if start_plot and end_plot:
                self.ax.set_xlim(xmin=start_plot, xmax=end_plot)
            else:
                self.ax.set_xlim(xmin=min(tmp_x), xmax=max(tmp_x))

    def x_formatter_cb(self, ax):
        if self.gdata.key_type == "string":
            smap = self.gdata.getStringMap()
            reverse_smap = {}
            for key, val in smap.items():
                reverse_smap[val] = key
            ticks = sorted(smap.values())
            ax.set_xticks([i + 0.5 for i in ticks])
            ax.set_xticklabels([reverse_smap[i] for i in ticks])
            labels = ax.get_xticklabels()
            ax.grid(False)
            if self.log_xaxis:
                xmin = 0.001
            else:
                xmin = 0
            ax.set_xlim(xmin=xmin, xmax=len(ticks))
        elif self.gdata.key_type == "time":
            dl = PrettyDateLocator()
            df = PrettyDateFormatter(dl)
            ax.xaxis.set_major_locator(dl)
            ax.xaxis.set_major_formatter(df)
            ax.xaxis.set_clip_on(False)
            sf = PrettyScalarFormatter()
            ax.yaxis.set_major_formatter(sf)

        else:
            return None



"""  The SiteSEMapping module performs the necessary CS gymnastics to
     resolve site and SE combinations.  These manipulations are necessary
     in several components.

     Assumes CS structure of: /Resources/Sites/<GRIDNAME>/<SITENAME>
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers, siteGridName


def getSEParameters(seName):
    """get all the SE parameters in a list

    :param str seName: name of the Storage Element

    :return: S_OK() with list of dict with parameters
    """
    # This import is here to avoid circular imports
    from DIRAC.Resources.Storage.StorageElement import StorageElement

    se = StorageElement(seName, hideExceptions=True)

    protocolsSet = set(se.localAccessProtocolList) | set(se.localWriteProtocolList)

    seParametersList = []
    for protocol in protocolsSet:
        seParameters = se.getStorageParameters(protocol=protocol)
        if seParameters["OK"]:
            seParametersList.append(seParameters["Value"])
        else:
            gLogger.verbose("No SE parameters obtained", "for SE %s and protocol %s" % (seName, protocol))

    return S_OK(seParametersList)


def getSEHosts(seName):
    """Get StorageElement host names (can be more than one depending on the protocol)

    :param str seName: name of the storage element

    :return: S_OK() with list of hosts or S_ERROR
    """

    seParameters = getSEParameters(seName)
    if not seParameters["OK"]:
        gLogger.warn("Could not get SE parameters", "SE: %s" % seName)
        return seParameters

    return S_OK([parameters["Host"] for parameters in seParameters["Value"]])


def getStorageElementsHosts(seNames=None):
    """Get StorageElement host names

    :param list seNames: possible list of storage element names (if not provided, will use all)
    :param list plugins: if provided, restrict to a certain list of plugins

    :return: S_OK() with list of hosts or S_ERROR
    """

    seHosts = []

    if seNames is None:
        seNames = DMSHelpers().getStorageElements()

    for seName in seNames:

        try:
            seHost = getSEHosts(seName)
            if not seHost["OK"]:
                gLogger.warn("Could not get SE Host", "SE: %s" % seName)
                continue
            if seHost["Value"]:
                seHosts.extend(seHost["Value"])
        except Exception as excp:  # pylint: disable=broad-except
            gLogger.error("Failed to get SE %s information (SE skipped) " % seName)
            gLogger.exception("Operation finished  with exception: ", lException=excp)
    return S_OK(list(set(seHosts)))


#############################################################################
def getSiteSEMapping(gridName="", withSiteLocalSEMapping=False):
    """Returns a dictionary of all sites and their localSEs as a list, e.g.
    {'LCG.CERN.ch':['CERN-RAW','CERN-RDST',...]}
    If gridName is specified, result is restricted to that Grid type.
    """
    result = DMSHelpers().getSiteSEMapping()
    if not result["OK"]:
        return result
    if withSiteLocalSEMapping:
        mapping = result["Value"][2]
    else:
        mapping = result["Value"][1]
    if gridName:
        mapping = dict((site, mapping[site]) for site in mapping if siteGridName(site) == gridName)
    return S_OK(mapping)


#############################################################################
def getSESiteMapping(gridName="", withSiteLocalSEMapping=False):
    """Returns a dictionary of all SEs and their associated site(s), e.g.
    {'CERN-RAW':'LCG.CERN.ch','CERN-RDST':'LCG.CERN.ch',...]}
    Although normally one site exists for a given SE, it is possible over all
    Grid types to have multiple entries.
    If gridName is specified, result is restricted to that Grid type.
    Assumes CS structure of: /Resources/Sites/<GRIDNAME>/<SITENAME>
    """
    storageElements = DMSHelpers().getStorageElements()
    return S_OK(
        dict(
            (se, getSitesForSE(se, gridName=gridName, withSiteLocalSEMapping=withSiteLocalSEMapping).get("Value", []))
            for se in storageElements
        )
    )


#############################################################################


def getSitesForSE(storageElement, gridName="", withSiteLocalSEMapping=False):
    """Given a DIRAC SE name this method returns a list of corresponding sites.
    Optionally restrict to Grid specified by name.
    """

    result = DMSHelpers().getSitesForSE(
        storageElement, connectionLevel="DOWNLOAD" if withSiteLocalSEMapping else "LOCAL"
    )
    if not result["OK"] or not gridName:
        return result

    return S_OK([site for site in result["Value"] if siteGridName(site) == gridName])


#############################################################################
def getSEsForSite(siteName, withSiteLocalSEMapping=False):
    """Given a DIRAC site name this method returns a list of corresponding SEs."""
    result = DMSHelpers().getSEsForSite(siteName, connectionLevel="DOWNLOAD" if withSiteLocalSEMapping else "LOCAL")
    if not result["OK"]:
        return S_OK([])
    return result


#############################################################################


def isSameSiteSE(se1, se2):
    """Check if the 2 SEs are at the same site"""
    dmsHelper = DMSHelpers()
    site1 = dmsHelper.getLocalSiteForSE(se1).get("Value")
    site2 = dmsHelper.getLocalSiteForSE(se2).get("Value")
    return site1 and site2 and site1 == site2


#############################################################################


def getSEsForCountry(country):
    """Determines the associated SEs from the country code"""
    return DMSHelpers().getSEsAtCountry(country)



"""
    FileCatalogClientBase is a base class for the clients of file catalog-like
    services built within the DIRAC framework.

    The class contains variables defining lists of implemented catalog methods
    READ_METHODS
    WRITE_METHODS
    NO_LFN_METHODS
    ADMIN_METHODS

    Those lists must be complemented in the derived classes to include specific
    implemented methods.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities.ReturnValues import S_OK
from DIRAC.Resources.Catalog.Utilities import checkCatalogArguments


@createClient("DataManagement/FileCatalog")
class FileCatalogClientBase(Client):
    """Client code to the DIRAC File Catalogue"""

    # Default mandatory methods having all-allowing implementation in the base class
    READ_METHODS = ["hasAccess", "exists", "getPathPermissions"]

    # List of methods that can be complemented in the derived classes
    WRITE_METHODS = []
    NO_LFN_METHODS = []
    ADMIN_METHODS = []

    def __init__(self, url=None, **kwargs):
        """Constructor function."""
        super(FileCatalogClientBase, self).__init__(**kwargs)
        if url:
            self.serverURL = url
        self.available = False

    def isOK(self, timeout=120):
        """Check that the service is OK"""
        if not self.available:
            rpcClient = self._getRPC(timeout=timeout)
            res = rpcClient.isOK()
            if not res["OK"]:
                self.available = False
            else:
                self.available = True
        return S_OK(self.available)

    #######################################################################################
    #  The following methods must be implemented in derived classes
    #######################################################################################

    @classmethod
    def getInterfaceMethods(cls):
        """Get the methods implemented by the File Catalog client

        :return tuple: ( read_methods_list, write_methods_list, nolfn_methods_list )
        """
        return (cls.READ_METHODS, cls.WRITE_METHODS, cls.NO_LFN_METHODS)

    @classmethod
    def hasCatalogMethod(cls, methodName):
        """Check of a method with the given name is implemented

        :param str methodName: the name of the method to check
        :return: boolean Flag True if the method is implemented
        """
        return methodName in (cls.READ_METHODS + cls.WRITE_METHODS + cls.NO_LFN_METHODS)

    ########################################################################################
    #  Some default methods to always return successfully if we do not care
    ########################################################################################
    @checkCatalogArguments
    def hasAccess(self, paths, _opType):
        """Default method: returns True for all paths and all actions

        :param lfn paths: has to be formatted this way :
                    { lfn : { se1 : pfn1, se2 : pfn2, ...}, ...}
        :param str _opType: dummy string

        :return: { successful : { lfn : [ ses ] } : failed : { lfn : { se : msg } } }
        """
        lfns = list(paths)
        return S_OK({"Failed": {}, "Successful": dict.fromkeys(lfns, True)})

    @checkCatalogArguments
    def exists(self, lfns):
        """Default method: returns False for all paths

        :param lfn paths: has to be formatted this way :
                    { lfn : { se1 : pfn1, se2 : pfn2, ...}, ...}

        :return: { successful : { lfn : [ ses ] } : failed : { lfn : { se : msg } } }

        """
        return S_OK({"Failed": {}, "Successful": dict.fromkeys(lfns, False)})

    @checkCatalogArguments
    def getPathPermissions(self, lfns):
        """Default method: returns Read & Write permission for all the paths

        :param lfn paths: has to be formatted this way :
                    { lfn : { se1 : pfn1, se2 : pfn2, ...}, ...}

        :return: { successful : { lfn : [ ses ] } : failed : { lfn : { se : msg } } }

        """
        return S_OK({"Failed": {}, "Successful": dict.fromkeys(lfns, {"Write": True, "Read": True})})



"""
Module containing functions interacting with the CS and useful for the RSS
modules.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gConfig, gLogger, S_OK
from DIRAC.Core.Utilities.SiteSEMapping import getSEParameters
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getQueues, getCESiteMapping
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.ResourceStatusSystem.Utilities import Utils


def warmUp():
    """
    gConfig has its own dark side, it needs some warm up phase.
    """
    from DIRAC.ConfigurationSystem.private.Refresher import gRefresher

    gRefresher.refreshConfigurationIfNeeded()


def getResources():
    """
    Gets all resources
    """

    resources = DMSHelpers().getStorageElements()

    fts = getFTS()
    if fts["OK"]:
        resources = resources + fts["Value"]

    fc = getFileCatalogs()
    if fc["OK"]:
        resources = resources + fc["Value"]

    res = getCESiteMapping()
    if res["OK"]:
        resources = resources + list(res["Value"])

    return S_OK(resources)


def getStorageElementEndpoint(seName):
    """Get endpoints of a StorageElement

    :param str seName: name of the storage element

    :returns: S_OK() or S_ERROR
              for historical reasons, if the protocol is SRM, you get  'httpg://host:port/WSUrl'
              For other protocols, you get :py:meth:`~DIRAC.Resources.Storage.StorageBase.StorageBase.getEndpoint`

    """
    seParameters = getSEParameters(seName)
    if not seParameters["OK"]:
        gLogger.warn("Could not get SE parameters", "for SE %s" % seName)
        return seParameters

    seEndpoints = []

    for parameters in seParameters["Value"]:
        if parameters["Protocol"].lower() == "srm":
            # we need to construct the URL with httpg://
            host = parameters["Host"]
            port = parameters["Port"]
            wsurl = parameters["WSUrl"]
            # MAYBE wusrl is not defined
            if host and port:
                url = "httpg://%s:%s%s" % (host, port, wsurl)
                url = url.replace("?SFN=", "")
                seEndpoints.append(url)
        else:
            seEndpoints.append(parameters["Endpoint"])

    return S_OK(seEndpoints)


def getFTS():
    """
    Gets all FTS endpoints
    """

    result = gConfig.getOptions("Resources/FTSEndpoints/FTS3")
    if result["OK"]:
        return result
    return S_OK([])


def getSpaceTokenEndpoints():
    """Get Space Token Endpoints"""

    return Utils.getCSTree("Shares/Disk")


def getFileCatalogs():
    """
    Gets all storage elements from /Resources/FileCatalogs
    """

    _basePath = "Resources/FileCatalogs"

    fileCatalogs = gConfig.getSections(_basePath)
    return fileCatalogs


def getSiteElements(siteName):
    """
    Gets all the computing and storage elements for a given site
    """

    res = DMSHelpers().getSiteSEMapping()
    if not res["OK"]:
        return res
    resources = res["Value"][1].get(siteName, [])

    res = getQueues(siteName)
    if not res["OK"]:
        return res
    resources = list(resources) + list(res["Value"][siteName])

    return S_OK(resources)


def getQueuesRSS():
    """
    Gets all queues from /Resources/Sites/<>/<>/CEs/<>/Queues
    """
    _basePath = "Resources/Sites"

    queues = []

    domainNames = gConfig.getSections(_basePath)
    if not domainNames["OK"]:
        return domainNames
    domainNames = domainNames["Value"]

    for domainName in domainNames:
        domainSites = gConfig.getSections("%s/%s" % (_basePath, domainName))
        if not domainSites["OK"]:
            return domainSites
        domainSites = domainSites["Value"]

        for site in domainSites:
            siteCEs = gConfig.getSections("%s/%s/%s/CEs" % (_basePath, domainName, site))
            if not siteCEs["OK"]:
                # return siteCEs
                gLogger.error(siteCEs["Message"])
                continue
            siteCEs = siteCEs["Value"]

            for siteCE in siteCEs:
                siteQueue = gConfig.getSections("%s/%s/%s/CEs/%s/Queues" % (_basePath, domainName, site, siteCE))
                if not siteQueue["OK"]:
                    # return siteQueue
                    gLogger.error(siteQueue["Message"])
                    continue
                siteQueue = siteQueue["Value"]

                queues.extend(siteQueue)

    # Remove duplicated ( just in case )
    queues = list(set(queues))

    return S_OK(queues)



########################################################################
# File :    PilotStatusAgent.py
# Author :  Stuart Paterson
########################################################################
"""  The Pilot Status Agent updates the status of the pilot jobs in the
     PilotAgents database.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.AccountingSystem.Client.Types.Pilot import Pilot as PilotAccounting
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities import Time
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
from DIRAC.WorkloadManagementSystem.Client import PilotStatus
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient
from DIRAC.WorkloadManagementSystem.DB.PilotAgentsDB import PilotAgentsDB
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB

MAX_JOBS_QUERY = 10
MAX_WAITING_STATE_LENGTH = 3


class PilotStatusAgent(AgentModule):
    """
    The specific agents must provide the following methods:
      - initialize() for initial settings
      - beginExecution()
      - execute() - the main method called in the agent cycle
      - endExecution()
      - finalize() - the graceful exit of the method, this one is usually used
                 for the agent restart
    """

    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)

        self.jobDB = None
        self.pilotDB = None
        self.diracadmin = None

    #############################################################################
    def initialize(self):
        """Sets defaults"""

        self.am_setOption("PollingTime", 120)
        self.am_setOption("GridEnv", "")
        self.am_setOption("PilotStalledDays", 3)
        self.pilotDB = PilotAgentsDB()
        self.diracadmin = DiracAdmin()
        self.jobDB = JobDB()
        self.clearPilotsDelay = self.am_getOption("ClearPilotsDelay", 30)
        self.clearAbortedDelay = self.am_getOption("ClearAbortedPilotsDelay", 7)
        self.pilots = PilotManagerClient()

        return S_OK()

    #############################################################################
    def execute(self):
        """The PilotAgent execution method."""

        self.pilotStalledDays = self.am_getOption("PilotStalledDays", 3)
        self.gridEnv = self.am_getOption("GridEnv")
        if not self.gridEnv:
            # No specific option found, try a general one
            setup = gConfig.getValue("/DIRAC/Setup", "")
            if setup:
                instance = gConfig.getValue("/DIRAC/Setups/%s/WorkloadManagement" % setup, "")
                if instance:
                    self.gridEnv = gConfig.getValue("/Systems/WorkloadManagement/%s/GridEnv" % instance, "")
        result = self.pilotDB._getConnection()
        if result["OK"]:
            connection = result["Value"]
        else:
            return result

        # Now handle pilots not updated in the last N days (most likely the Broker is no
        # longer available) and declare them Deleted.
        result = self.handleOldPilots(connection)

        connection.close()

        result = self.pilots.clearPilots(self.clearPilotsDelay, self.clearAbortedDelay)
        if not result["OK"]:
            self.log.warn("Failed to clear old pilots in the PilotAgentsDB")

        return S_OK()

    def handleOldPilots(self, connection):
        """
        select all pilots that have not been updated in the last N days and declared them
        Deleted, accounting for them.
        """
        pilotsToAccount = {}
        timeLimitToConsider = Time.toString(Time.dateTime() - Time.day * self.pilotStalledDays)
        result = self.pilotDB.selectPilots(
            {"Status": PilotStatus.PILOT_TRANSIENT_STATES}, older=timeLimitToConsider, timeStamp="LastUpdateTime"
        )
        if not result["OK"]:
            self.log.error("Failed to get the Pilot Agents")
            return result
        if not result["Value"]:
            return S_OK()

        refList = result["Value"]
        result = self.pilotDB.getPilotInfo(refList)
        if not result["OK"]:
            self.log.error("Failed to get Info for Pilot Agents")
            return result

        pilotsDict = result["Value"]

        for pRef in pilotsDict:
            if pilotsDict[pRef].get("Jobs") and self._checkJobLastUpdateTime(
                pilotsDict[pRef]["Jobs"], self.pilotStalledDays
            ):
                self.log.debug(
                    "%s should not be deleted since one job of %s is running."
                    % (str(pRef), str(pilotsDict[pRef]["Jobs"]))
                )
                continue
            deletedJobDict = pilotsDict[pRef]
            deletedJobDict["Status"] = PilotStatus.DELETED
            deletedJobDict["StatusDate"] = Time.dateTime()
            pilotsToAccount[pRef] = deletedJobDict
            if len(pilotsToAccount) > 100:
                self.accountPilots(pilotsToAccount, connection)
                self._killPilots(pilotsToAccount)
                pilotsToAccount = {}

        self.accountPilots(pilotsToAccount, connection)
        self._killPilots(pilotsToAccount)

        return S_OK()

    def accountPilots(self, pilotsToAccount, connection):
        """account for pilots"""
        accountingFlag = False
        pae = self.am_getOption("PilotAccountingEnabled", "yes")
        if pae.lower() == "yes":
            accountingFlag = True

        if not pilotsToAccount:
            self.log.info("No pilots to Account")
            return S_OK()

        accountingSent = False
        if accountingFlag:
            retVal = self.pilotDB.getPilotInfo(list(pilotsToAccount), conn=connection)
            if not retVal["OK"]:
                self.log.error("Fail to retrieve Info for pilots", retVal["Message"])
                return retVal
            dbData = retVal["Value"]
            for pref in dbData:
                if pref in pilotsToAccount:
                    if dbData[pref]["Status"] not in PilotStatus.PILOT_FINAL_STATES:
                        dbData[pref]["Status"] = pilotsToAccount[pref]["Status"]
                        dbData[pref]["DestinationSite"] = pilotsToAccount[pref]["DestinationSite"]
                        dbData[pref]["LastUpdateTime"] = pilotsToAccount[pref]["StatusDate"]

            retVal = self._addPilotsAccountingReport(dbData)
            if not retVal["OK"]:
                self.log.error("Fail to retrieve Info for pilots", retVal["Message"])
                return retVal

            self.log.info("Sending accounting records...")
            retVal = gDataStoreClient.commit()
            if not retVal["OK"]:
                self.log.error("Can't send accounting reports", retVal["Message"])
            else:
                self.log.info("Accounting sent for %s pilots" % len(pilotsToAccount))
                accountingSent = True

        if not accountingFlag or accountingSent:
            for pRef in pilotsToAccount:
                pDict = pilotsToAccount[pRef]
                self.log.verbose("Setting Status for %s to %s" % (pRef, pDict["Status"]))
                self.pilotDB.setPilotStatus(
                    pRef, pDict["Status"], pDict["DestinationSite"], pDict["StatusDate"], conn=connection
                )

        return S_OK()

    def _addPilotsAccountingReport(self, pilotsData):
        """fill accounting data"""
        for pRef in pilotsData:
            pData = pilotsData[pRef]
            pA = PilotAccounting()
            pA.setEndTime(pData["LastUpdateTime"])
            pA.setStartTime(pData["SubmissionTime"])
            retVal = Registry.getUsernameForDN(pData["OwnerDN"])
            if not retVal["OK"]:
                userName = "unknown"
                self.log.error(
                    "Can't determine username for dn",
                    ": %s : %s" % (pData["OwnerDN"], retVal["Message"]),
                )
            else:
                userName = retVal["Value"]
            pA.setValueByKey("User", userName)
            pA.setValueByKey("UserGroup", pData["OwnerGroup"])
            result = getCESiteMapping(pData["DestinationSite"])
            if result["OK"] and pData["DestinationSite"] in result["Value"]:
                pA.setValueByKey("Site", result["Value"][pData["DestinationSite"]].strip())
            else:
                pA.setValueByKey("Site", "Unknown")
            pA.setValueByKey("GridCE", pData["DestinationSite"])
            pA.setValueByKey("GridMiddleware", pData["GridType"])
            pA.setValueByKey("GridResourceBroker", pData["Broker"])
            pA.setValueByKey("GridStatus", pData["Status"])
            if "Jobs" not in pData:
                pA.setValueByKey("Jobs", 0)
            else:
                pA.setValueByKey("Jobs", len(pData["Jobs"]))
            self.log.verbose("Added accounting record for pilot %s" % pData["PilotID"])
            retVal = gDataStoreClient.addRegister(pA)
            if not retVal["OK"]:
                return retVal
        return S_OK()

    def _killPilots(self, acc):
        for i in sorted(acc.keys()):
            result = self.diracadmin.getPilotInfo(i)
            if result["OK"] and i in result["Value"] and "Status" in result["Value"][i]:
                ret = self.diracadmin.killPilot(str(i))
                if ret["OK"]:
                    self.log.info("Successfully deleted", ": %s (Status : %s)" % (i, result["Value"][i]["Status"]))
                else:
                    self.log.error("Failed to delete pilot: ", "%s : %s" % (i, ret["Message"]))
            else:
                self.log.error("Failed to get pilot info", "%s : %s" % (i, str(result)))

    def _checkJobLastUpdateTime(self, joblist, StalledDays):
        timeLimitToConsider = Time.dateTime() - Time.day * StalledDays
        ret = False
        for jobID in joblist:
            result = self.jobDB.getJobAttributes(int(jobID))
            if result["OK"]:
                if "LastUpdateTime" in result["Value"]:
                    lastUpdateTime = result["Value"]["LastUpdateTime"]
                    if Time.fromString(lastUpdateTime) > timeLimitToConsider:
                        ret = True
                        self.log.debug(
                            "Since %s updates LastUpdateTime on %s this does not to need to be deleted."
                            % (str(jobID), str(lastUpdateTime))
                        )
                        break
            else:
                self.log.error("Error taking job info from DB", result["Message"])
        return ret



#!/usr/bin/env python
"""
Create a new DB in the MySQL server
"""
# Script initialization and parseCommandLine
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["DB: Name of the Database"])
    _, args = Script.parseCommandLine()

    # Script imports
    from DIRAC import gConfig
    from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller
    from DIRAC.FrameworkSystem.Utilities import MonitoringUtilities

    gComponentInstaller.exitOnError = True
    gComponentInstaller.getMySQLPasswords()
    for db in args:
        result = gComponentInstaller.installDatabase(db)
        if not result["OK"]:
            print("ERROR: failed to correctly install %s" % db, result["Message"])
            continue
        extension, system = result["Value"]
        gComponentInstaller.addDatabaseOptionsToCS(gConfig, system, db, overwrite=True)

        if db != "InstalledComponentsDB":
            result = MonitoringUtilities.monitorInstallation("DB", system, db)
            if not result["OK"]:
                print("ERROR: failed to register installation in database: %s" % result["Message"])


if __name__ == "__main__":
    main()



#! /usr/bin/env python
########################################################################
# File :    dirac-stager-monitor-requests
# Author :  Daniela Remenska
########################################################################
"""
Report the details of file staging requests, based on selection filters

WARNING: Query may be heavy, please use --limit switch!

Example:
  $ dirac-stager-show-requests.py --status=Staged --se=GRIDKA-RDST --limit=10 --showJobs=YES
  Query limited to 10 entries

  Status          LastUpdate           LFN              SE       Reason     Jobs       PinExpiryTime        PinLength
  Staged   2013-06-05 20:10:50 /lhcb/LHCb/5.full.dst GRIDKA-RDST None    ['48498752']  2013-06-05 22:10:50  86400
  Staged   2013-06-06 15:54:29 /lhcb/LHCb/1.full.dst GRIDKA-RDST None    ['48516851']  2013-06-06 16:54:29  43200
  Staged   2013-06-07 02:35:41 /lhcb/LHCb/3.full.dst GRIDKA-RDST None    ['48520736']  2013-06-07 03:35:41  43200
  Staged   2013-06-06 04:16:50 /lhcb/LHCb/3.full.dst GRIDKA-RDST None    ['48510852']  2013-06-06 06:16:50  86400
  Staged   2013-06-07 03:44:04 /lhcb/LHCb/3.full.dst GRIDKA-RDST None    ['48520737']  2013-06-07 04:44:04  43200
  Staged   2013-06-05 23:37:46 /lhcb/LHCb/3.full.dst GRIDKA-RDST None    ['48508687']  2013-06-06 01:37:46  86400
  Staged   2013-06-10 08:50:09 /lhcb/LHCb/5.full.dst GRIDKA-RDST None    ['48518896']  2013-06-10 09:50:09  43200
  Staged   2013-06-06 11:03:25 /lhcb/LHCb/2.full.dst GRIDKA-RDST None    ['48515583']  2013-06-06 12:03:25  43200
  Staged   2013-06-06 11:11:50 /lhcb/LHCb/2.full.dst GRIDKA-RDST None    ['48515072']  2013-06-06 12:11:50  43200
  Staged   2013-06-07 03:19:26 /lhcb/LHCb/2.full.dst GRIDKA-RDST None    ['48515600']  2013-06-07 04:19:26  43200
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC import gConfig, gLogger, exit as DIRACExit, S_OK, version

subLogger = None


@Script()
def main():
    def registerSwitches():
        """
        Registers all switches that can be used while calling the script from the
        command line interface.
        """

        switches = (
            (
                "status=",
                "Filter per file status=(New, Offline, Waiting, Failed, StageSubmitted, Staged)."
                "\n                                 If not used, all status values will be taken into account",
            ),
            ("se=", "Filter per Storage Element. If not used, all storage elements will be taken into account."),
            ("limit=", "Limit the number of entries returned."),
            ("showJobs=", "Whether to ALSO list the jobs asking for these files to be staged"),
        )

        for switch in switches:
            Script.registerSwitch("", switch[0], switch[1])

    def parseSwitches():
        """
        Parses the arguments passed by the user
        """

        Script.parseCommandLine(ignoreErrors=True)
        args = Script.getPositionalArgs()
        if args:
            subLogger.error("Found the following positional args '%s', but we only accept switches" % args)
            subLogger.error("Please, check documentation below")
            Script.showHelp(exitCode=1)

        switches = dict(Script.getUnprocessedSwitches())

        for key in ("status", "se", "limit"):
            if key not in switches:
                subLogger.warn("You're not using switch --%s, query may take long!" % key)

        if "status" in switches and switches["status"] not in (
            "New",
            "Offline",
            "Waiting",
            "Failed",
            "StageSubmitted",
            "Staged",
        ):
            subLogger.error('Found "%s" as Status value. Incorrect value used!' % switches["status"])
            subLogger.error("Please, check documentation below")
            Script.showHelp(exitCode=1)

        subLogger.debug("The switches used are:")
        map(subLogger.debug, switches.items())

        return switches

    # ...............................................................................

    def run():
        global subLogger

        from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient

        client = StorageManagerClient()
        queryDict = {}

        if "status" in switchDict:
            queryDict["Status"] = str(switchDict["status"])

        if "se" in switchDict:
            queryDict["SE"] = str(switchDict["se"])

        # weird: if there are no switches (dictionary is empty), then the --limit is ignored!!
        # must FIX that in StorageManagementDB.py!
        # ugly fix:
        newer = "1903-08-02 06:24:38"  # select newer than
        if "limit" in switchDict:
            gLogger.notice("Query limited to %s entries" % switchDict["limit"])
            res = client.getCacheReplicas(queryDict, None, newer, None, None, int(switchDict["limit"]))
        else:
            res = client.getCacheReplicas(queryDict)

        if not res["OK"]:
            gLogger.error(res["Message"])
        outStr = "\n"
        if res["Records"]:
            replicas = res["Value"]
            outStr += " %s" % ("Status".ljust(15))
            outStr += " %s" % ("LastUpdate".ljust(20))
            outStr += " %s" % ("LFN".ljust(80))
            outStr += " %s" % ("SE".ljust(10))
            outStr += " %s" % ("Reason".ljust(10))
            if "showJobs" in switchDict:
                outStr += " %s" % ("Jobs".ljust(10))
            outStr += " %s" % ("PinExpiryTime".ljust(15))
            outStr += " %s" % ("PinLength(sec)".ljust(15))
            outStr += "\n"

            for crid, info in replicas.items():
                outStr += " %s" % (info["Status"].ljust(15))
                outStr += " %s" % (str(info["LastUpdate"]).ljust(20))
                outStr += " %s" % (info["LFN"].ljust(30))
                outStr += " %s" % (info["SE"].ljust(15))
                outStr += " %s" % (str(info["Reason"]).ljust(10))

                # Task info
                if "showJobs" in switchDict:
                    resTasks = client.getTasks({"ReplicaID": crid})
                    if resTasks["OK"]:
                        if resTasks["Value"]:
                            tasks = resTasks["Value"]
                            jobs = []
                            for tid in tasks:
                                jobs.append(tasks[tid]["SourceTaskID"])
                            outStr += " %s " % (str(jobs).ljust(10))
                    else:
                        outStr += " %s " % (" --- ".ljust(10))
                # Stage request info
                # what if there's no request to the site yet?
                resStageRequests = client.getStageRequests({"ReplicaID": crid})
                if not resStageRequests["OK"]:
                    gLogger.error(resStageRequests["Message"])
                if resStageRequests["Records"]:
                    stageRequests = resStageRequests["Value"]
                    for info in stageRequests.values():
                        outStr += " %s" % (str(info["PinExpiryTime"]).ljust(20))
                        outStr += " %s" % (str(info["PinLength"]).ljust(10))
                outStr += "\n"

            gLogger.notice(outStr)
        else:
            gLogger.notice("No entries")

    subLogger = gLogger.getSubLogger(__file__)

    registerSwitches()

    switchDict = parseSwitches()
    run()

    # Bye
    DIRACExit(0)


if __name__ == "__main__":
    main()



"""
Utilities for ComponentMonitoring features
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import socket

from DIRAC import S_OK
from DIRAC.FrameworkSystem.Client.ComponentMonitoringClient import ComponentMonitoringClient
from DIRAC.Core.Security.ProxyInfo import getProxyInfo


def monitorInstallation(componentType, system, component, module=None, cpu=None, hostname=None):
    """
    Register the installation of a component in the ComponentMonitoringDB
    """
    monitoringClient = ComponentMonitoringClient()

    if not module:
        module = component

    # Retrieve user installing the component
    user = None
    result = getProxyInfo()
    if result["OK"]:
        proxyInfo = result["Value"]
        if "username" in proxyInfo:
            user = proxyInfo["username"]
    else:
        return result
    if not user:
        user = "unknown"

    if not cpu:
        cpu = "Not available"
        for line in open("/proc/cpuinfo"):
            if line.startswith("model name"):
                cpu = line.split(":")[1][0:64]
                cpu = cpu.replace("\n", "").lstrip().rstrip()

    if not hostname:
        hostname = socket.getfqdn()
    instance = component[0:32]

    result = monitoringClient.installationExists(
        {"Instance": instance, "UnInstallationTime": None},
        {"Type": componentType, "DIRACSystem": system, "DIRACModule": module},
        {"HostName": hostname, "CPU": cpu},
    )

    if not result["OK"]:
        return result
    if result["Value"]:
        return S_OK("Monitoring of %s is already enabled" % component)

    result = monitoringClient.addInstallation(
        {"InstallationTime": datetime.datetime.utcnow(), "InstalledBy": user, "Instance": instance},
        {"Type": componentType, "DIRACSystem": system, "DIRACModule": module},
        {"HostName": hostname, "CPU": cpu},
        True,
    )
    return result


def monitorUninstallation(system, component, cpu=None, hostname=None):
    """
    Register the uninstallation of a component in the ComponentMonitoringDB
    """
    monitoringClient = ComponentMonitoringClient()

    # Retrieve user uninstalling the component
    user = None
    result = getProxyInfo()
    if result["OK"]:
        proxyInfo = result["Value"]
        if "username" in proxyInfo:
            user = proxyInfo["username"]
    else:
        return result
    if not user:
        user = "unknown"

    if not cpu:
        cpu = "Not available"
        for line in open("/proc/cpuinfo"):
            if line.startswith("model name"):
                cpu = line.split(":")[1][0:64]
                cpu = cpu.replace("\n", "").lstrip().rstrip()

    if not hostname:
        hostname = socket.getfqdn()
    instance = component[0:32]

    result = monitoringClient.updateInstallations(
        {"Instance": instance, "UnInstallationTime": None},
        {"DIRACSystem": system},
        {"HostName": hostname, "CPU": cpu},
        {"UnInstallationTime": datetime.datetime.utcnow(), "UnInstalledBy": user},
    )
    return result



""" WMSHistoryCorrector is the implementation class of the BaseHistoryCorrector
    which gets resources consumption history data from the WMSHistory Accounting
    database
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime

from DIRAC import gLogger, S_OK

from DIRAC.WorkloadManagementSystem.private.correctors.BaseHistoryCorrector import BaseHistoryCorrector
from DIRAC.Core.Utilities import Time
from DIRAC.AccountingSystem.Client.ReportsClient import ReportsClient


class WMSHistoryCorrector(BaseHistoryCorrector):
    def initialize(self):
        super(WMSHistoryCorrector, self).initialize()
        self.log = gLogger.getSubLogger("WMSHistoryCorrector")
        return S_OK()

    def _getHistoryData(self, timeSpan, groupToUse):
        """Get history data from Accounting WMSHistory database

        :param int timeSpan: time span
        :param str groupToUse: requested user group
        :return: dictionary with history data
        """
        reportsClient = ReportsClient()

        reportCondition = {"Status": ["Running"]}
        if not groupToUse:
            reportGrouping = "UserGroup"
        else:
            reportGrouping = "User"
            reportCondition = {"UserGroup": groupToUse}
        now = Time.dateTime()
        result = reportsClient.getReport(
            "WMSHistory",
            "AverageNumberOfJobs",
            now - datetime.timedelta(seconds=timeSpan),
            now,
            reportCondition,
            reportGrouping,
            {"lastSeconds": timeSpan},
        )
        return result



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# List of OIDs used in handling VOMS extension.
# VOMS extension is encoded in ASN.1 format and it's surprisingly hard to decode. OIDs describe content of sections
# of the data. There is no "official list of OIDs", ones used here are sourced from analyzing VOMS extensions itself
# and different pieces of code and presentations in subject of X509 certificates, certificate extensions and VOMS.
# Googling names or values of those OIDs, especially VOMS related, usually result in up to three pages of results,
# mainly Java code defining those values like code below.
# This is literally lookup table, so I know WTH is this, when I read value and see '1.3.6.1.4.1.8005.100.100.4'.

DOMAIN_COMPONENT_OID = "0.9.2342.19200300.100.1.25"
DIRAC_GROUP_OID = "1.2.42.42"
VOMS_FQANS_OID = "1.3.6.1.4.1.8005.100.100.4"
VOMS_EXTENSION_OID = "1.3.6.1.4.1.8005.100.100.5"
VOMS_TAGS_EXT_OID = "1.3.6.1.4.1.8005.100.100.11"
COMMON_NAME_OID = "2.5.4.3"
SURNAME_OID = "2.5.4.4"
SERIALNUMBER_OID = "2.5.4.5"
COUNTRY_NAME = "2.5.4.6"
LOCALITY_NAME = "2.5.4.7"
STATE_OR_PROVINCE_NAME = "2.5.4.8"
ORGANIZATION_NAME = "2.5.4.10"
ORGANIZATIONAL_UNIT_NAME_OID = "2.5.4.11"
TITLE_OID = "2.5.4.12"
GIVEN_NAME_OID = "2.5.4.42"


# See https://tools.ietf.org/html/rfc3820#appendix-A
PROXY_OID = "1.3.6.1.5.5.7.21.1"
LIMITED_PROXY_OID = "1.3.6.1.4.1.3536.1.1.1.9"

# Some specific distinguished names: https://www.cryptosys.net/pki/manpki/pki_distnames.html

DN_MAPPING = {
    COMMON_NAME_OID: "/CN=",
    COUNTRY_NAME: "/C=",
    DOMAIN_COMPONENT_OID: "/DC=",
    GIVEN_NAME_OID: "/G=",
    LOCALITY_NAME: "/L=",
    ORGANIZATION_NAME: "/O=",
    ORGANIZATIONAL_UNIT_NAME_OID: "/OU=",
    SERIALNUMBER_OID: "/SERIALNUMBER=",
    STATE_OR_PROVINCE_NAME: "/ST=",
    SURNAME_OID: "/SN=",
    TITLE_OID: "/T=",
}



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from DIRAC import gLogger
from DIRAC.Core.Base.Client import Client, createClient


@createClient("WorkloadManagement/VirtualMachineManager")
class VMClient(Client):
    def __init__(self, *args, **kwargs):
        super(VMClient, self).__init__(*args, **kwargs)
        self.log = gLogger.getSubLogger("WorkloadManagement/VMClient")
        self.setServer("WorkloadManagement/VirtualMachineManager")



########################################################################
# File :   CSCLI.py
# Author : Adria Casajus
########################################################################

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import sys
import atexit
import os
import readline
import six

from DIRAC import gLogger

from DIRAC.Core.Utilities.File import mkDir
from DIRAC.Core.Base.CLI import CLI, colorize
from DIRAC.ConfigurationSystem.private.Modificator import Modificator
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient

__RCSID__ = "$Id$"


def _showTraceback():
    import traceback

    excepType, execpValue = sys.exc_info()[:2]
    print("________________________\n")
    print("Exception", excepType, ":", execpValue)
    traceback.print_tb(sys.exc_info()[2])
    print("________________________\n")


def _printComment(comment):
    commentList = comment.split("\n")
    for commentLine in commentList[:-1]:
        print("# %s" % commentLine.strip())


def _appendExtensionIfMissing(filename):
    dotPosition = filename.rfind(".")
    if dotPosition > -1:
        filename = filename[:dotPosition]
    return "%s.cfg" % filename


class CSCLI(CLI):
    def __init__(self):
        CLI.__init__(self)
        self.connected = False
        self.masterURL = "unset"
        self.writeEnabled = False
        self.modifiedData = False
        self.rpcClient = None
        self.do_connect()
        if self.connected:
            self.modificator = Modificator(self.rpcClient)
        else:
            self.modificator = Modificator()
        self.indentSpace = 20
        self.backupFilename = "dataChanges"
        # store history
        histfilename = os.path.basename(sys.argv[0])
        historyFile = os.path.expanduser("~/.dirac/%s.history" % histfilename[0:-3])
        mkDir(os.path.dirname(historyFile))
        if os.path.isfile(historyFile):
            readline.read_history_file(historyFile)
        readline.set_history_length(1000)
        atexit.register(readline.write_history_file, historyFile)

    def start(self):
        if self.connected:
            self.modificator.loadFromRemote()
            retVal = self.modificator.loadCredentials()
            if not retVal["OK"]:
                print("There was an error gathering your credentials")
                print(retVal["Message"])
                self._setStatus(False)
        try:
            self.cmdloop()
        except KeyboardInterrupt:
            gLogger.warn("Received a keyboard interrupt.")
            self.do_quit("")

    def _setConnected(self, connected, writeEnabled):
        self.connected = connected
        self.modifiedData = False
        self.writeEnabled = writeEnabled
        if connected:
            if writeEnabled:
                self.prompt = "(%s)-%s> " % (self.masterURL, colorize("Connected", "green"))
            else:
                self.prompt = "(%s)-%s> " % (self.masterURL, colorize("Connected (RO)", "yellow"))
        else:
            self.prompt = "(%s)-%s> " % (self.masterURL, colorize("Disconnected", "red"))

    def do_quit(self, dummy):
        """
        Exits the application without sending changes to server

        Usage: quit
        """
        print()
        if self.modifiedData:
            print("Changes are about to be written to file for later use.")
            self.do_writeToFile(self.backupFilename)
            print("Changes written to %s.cfg" % self.backupFilename)
        sys.exit(0)

    def _setStatus(self, connected=True):
        if not connected:
            self.masterURL = "unset"
            self._setConnected(False, False)
        else:
            retVal = self.rpcClient.writeEnabled()
            if retVal["OK"]:
                if retVal["Value"]:
                    self._setConnected(True, True)
                else:
                    self._setConnected(True, False)
            else:
                print("Server returned an error: %s" % retVal["Message"])
                self._setConnected(True, False)

    def _tryConnection(self):
        print("Trying connection to %s" % self.masterURL)
        try:
            self.rpcClient = ConfigurationClient(url=self.masterURL)
            self._setStatus()
        except Exception as x:
            gLogger.error("Couldn't connect to master CS server", "%s (%s)" % (self.masterURL, str(x)))
            self._setStatus(False)

    def do_connect(self, args=""):
        """
        Connects to configuration master server (in specified url if provided).

        Usage: connect <url>
        """
        if not args or not isinstance(args, six.string_types):
            self.masterURL = gConfigurationData.getMasterServer()
            if self.masterURL != "unknown" and self.masterURL:
                self._tryConnection()
            else:
                self._setStatus(False)
        else:
            splitted = args.split()
            if len(splitted) == 0:
                print("Must specify witch url to connect")
                self._setStatus(False)
            else:
                self.masterURL = splitted[0].strip()
                self._tryConnection()

    def do_sections(self, args):
        """
        Shows all sections with their comments.
        If no section is specified, root is taken.

        Usage: sections <section>
        """
        try:
            argList = args.split()
            if argList:
                baseSection = argList[0].strip()
            else:
                baseSection = "/"
            if not self.modificator.existsSection(baseSection):
                print("Section %s does not exist" % baseSection)
                return
            sectionList = self.modificator.getSections(baseSection)
            if not sectionList:
                print("Section %s is empty" % baseSection)
                return
            for section in sectionList:
                section = "%s/%s" % (baseSection, section)
                self.printPair(section, self.modificator.getComment(section), " #")
        except Exception:
            _showTraceback()

    def do_options(self, args):
        """
        Shows all options and values of a specified section

        Usage: options <section>
        """
        try:
            argList = args.split()
            if argList:
                section = argList[0].strip()
            else:
                print("Which section?")
                return
            if not self.modificator.existsSection(section):
                print("Section %s does not exist" % section)
                return
            optionsList = self.modificator.getOptions(section)
            if not optionsList:
                print("Section %s has no options" % section)
                return
            for option in optionsList:
                _printComment(self.modificator.getComment("%s/%s" % (section, option)))
                self.printPair(option, self.modificator.getValue("%s/%s" % (section, option)), "=")
        except Exception:
            _showTraceback()

    def do_get(self, args):
        """
        Shows value and comment for specified option in section

        Usage: get <path to option>
        """
        try:
            argList = args.split()
            if argList:
                optionPath = argList[0].strip()
            else:
                print("Which option?")
                return
            if self.modificator.existsOption(optionPath):
                option = optionPath.split("/")[-1]
                _printComment(self.modificator.getComment(optionPath))
                self.printPair(option, self.modificator.getValue(optionPath), "=")
            else:
                print("Option %s does not exist" % optionPath)
        except Exception:
            _showTraceback()

    def do_writeToServer(self, dummy):
        """
        Sends changes to server.

        Usage: writeToServer
        """
        if not self.connected:
            print("You are not connected!")
            return
        try:
            if not self.writeEnabled:
                print("This server can't receive data modifications")
                return
            if not self.modifiedData:
                while True:
                    choice = six.moves.input(
                        "Data has not been modified, do you still want to upload changes? yes/no [no]: "
                    )
                    choice = choice.lower()
                    if choice in ("yes", "y"):
                        break
                    else:
                        print("Commit aborted")
                        return

            choice = six.moves.input("Do you really want to send changes to server? yes/no [no]: ")
            choice = choice.lower()
            if choice in ("yes", "y"):
                print("Uploading changes to %s (It may take some seconds)..." % self.masterURL)
                response = self.modificator.commit()
                if response["OK"]:
                    self.modifiedData = False
                    print("Data sent to server.")
                    self.modificator.loadFromRemote()
                else:
                    print("Error sending data, server said: %s" % response["Message"])
                return
            else:
                print("Commit aborted")
        except Exception as x:
            _showTraceback()
            print("Could not upload changes. ", str(x))

    def do_set(self, args):
        """
        Sets option's value

        Usage: set <optionPath> <value>...

        From second argument until the last one is considered option's value

        NOTE: If specified section does not exist it is created.
        """
        try:
            argsList = args.split()
            if len(argsList) < 2:
                print("Must specify option and value to use")
                return
            optionPath = argsList[0].strip()
            value = " ".join(argsList[1:]).strip()
            self.modificator.setOptionValue(optionPath, value)
            self.modifiedData = True
        except Exception as x:
            print("Cannot insert value: ", str(x))

    def do_removeOption(self, args):
        """
        Removes an option.

        Usage: removeOption <option>

        There can be empty sections.
        """
        try:
            argsList = args.split()
            if len(argsList) < 1:
                print("Must specify option to delete")
                return
            optionPath = argsList[0].strip()
            choice = six.moves.input("Are you sure you want to delete %s? yes/no [no]: " % optionPath)
            choice = choice.lower()
            if choice in ("yes", "y", "true"):
                if self.modificator.removeOption(optionPath):
                    self.modifiedData = True
                else:
                    print("Can't be deleted")
            else:
                print("Aborting removal.")
        except Exception as x:
            print("Error removing option, %s" % str(x))

    def do_removeSection(self, args):
        """
        Removes a section.

        Usage: removeSection <section>
        """
        try:
            argsList = args.split()
            if len(argsList) < 1:
                print("Must specify section to delete")
                return
            section = argsList[0].strip()
            choice = six.moves.input("Are you sure you want to delete %s? yes/no [no]: " % section)
            choice = choice.lower()
            if choice in ("yes", "y", "true"):
                if self.modificator.removeSection(section):
                    self.modifiedData = True
                else:
                    print("Can't be deleted")
            else:
                print("Aborting removal.")
        except Exception as x:
            print("Error removing section, %s" % str(x))

    def do_setComment(self, args):
        """
        Sets option or section's comment. Requested entry MUST exist.

        Usage: set <option/section> <comment>...

        From third argument until the last one is considered option's comment.
        """
        try:
            argsList = args.split()
            if len(argsList) < 2:
                print("Must specify option and value to use")
                return
            entryPath = argsList[0].strip()
            value = " ".join(argsList[1:]).strip()
            self.modificator.setComment(entryPath, value)
            self.modifiedData = True
        except Exception as x:
            print("Cannot insert comment: ", str(x))

    def do_writeToFile(self, args):
        """
        Writes modification to file for later use.

        Usage: writeToFile <filename>.cfg

        Note that if a file extension is specified, it is replaced by .cfg suffix.
        If not it is added automatically
        """
        try:
            if len(args) == 0:
                print("Filename to write must be specified!")
                return
            filename = args.split()[0].strip()
            filename = _appendExtensionIfMissing(filename)
            self.modificator.dumpToFile(filename)
        except Exception as x:
            print("Couldn't write to file %s: %s" % (filename, str(x)))

    def do_readFromFile(self, args):
        """
        Reads data from filename to be used. Actual data will be replaced!

        Usage: readFromFile <filename>.cfg

        Note that if a file extension is specified, it is replaced by .cfg suffix.
        If not it is added automatically
        """
        try:
            if len(args) == 0:
                print("Filename to read must be specified!")
                return
            filename = args.split()[0].strip()
            filename = _appendExtensionIfMissing(filename)
            self.modificator.loadFromFile(filename)
            self.modifiedData = True
        except Exception as x:
            print("Couldn't read from file %s: %s" % (filename, str(x)))

    def do_mergeFromFile(self, args):
        """
        Reads data from filename and merges it with current data.
        Data read from file has more precedence that current one.

        Usage: mergeFromFile <filename>.cfg

        Note that if a file extension is specified, it is replaced by .cfg suffix.
        If not it is added automatically
        """
        try:
            if len(args) == 0:
                print("Filename to read must be specified!")
                return
            filename = args.split()[0].strip()
            filename = _appendExtensionIfMissing(filename)
            self.modificator.mergeFromFile(filename)
            self.modifiedData = True
        except Exception as x:
            _showTraceback()
            print("Couldn't read from file %s: %s" % (filename, str(x)))

    def do_showData(self, dummy):
        """
        Shows the current modified configuration
        Usage: showData
        """
        print(self.modificator)

    def do_showHistory(self, args):
        """
        Shows the last commit history
        Usage: showHistory <update limit>
        """
        try:
            argsList = args.split()
            limit = 100
            if len(argsList) > 0:
                limit = int(argsList[0])
            history = self.modificator.getHistory(limit)
            print("%s recent commits:" % limit)
            for entry in history:
                self.printPair(entry[0], entry[1], "@")
        except Exception:
            _showTraceback()

    def do_showDiffWithServer(self, dummy):
        """
        Shows diff with lastest version in server
        Usage: showDiffWithServer
        """
        try:
            diffData = self.modificator.showCurrentDiff()
            print("Diff with latest from server ( + local - remote )")
            for line in diffData:
                if line[0] in ("-"):
                    print(colorize(line, "red"))
                elif line[0] in ("+"):
                    print(colorize(line, "green"))
                elif line[0] in ("?"):
                    print(colorize(line, "yellow"), end=" ")
        except Exception:
            _showTraceback()

    def do_showDiffBetweenVersions(self, args):
        """
        Shows diff between two versions
        Usage: showDiffBetweenVersions <version 1 with spaces> <version 2 with spaces>
        """
        try:
            argsList = args.split()
            if len(argsList) < 4:
                print("What are the two versions to compare?")
                return
            v1 = " ".join(argsList[0:2])
            v2 = " ".join(argsList[2:4])
            print("Comparing '%s' with '%s' " % (v1, v2))
            diffData = self.modificator.getVersionDiff(v1, v2)
            print("Diff with latest from server ( + %s - %s )" % (v2, v1))
            for line in diffData:
                if line[0] in ("-"):
                    print(colorize(line, "red"))
                elif line[0] in ("+"):
                    print(colorize(line, "green"))
                elif line[0] in ("?"):
                    print(colorize(line, "yellow"), end=" ")
                else:
                    print(line)
        except Exception:
            _showTraceback()

    def do_rollbackToVersion(self, args):
        """
        rolls back to user selected version of the configuration
        Usage: rollbackToVersion <version with spaces>>
        """
        try:
            argsList = args.split()
            if len(argsList) < 2:
                print("What version to rollback?")
                return
            version = " ".join(argsList[0:2])
            choice = six.moves.input("Do you really want to rollback to version %s? yes/no [no]: " % version)
            choice = choice.lower()
            if choice in ("yes", "y"):
                response = self.modificator.rollbackToVersion(version)
                if response["OK"]:
                    self.modifiedData = False
                    print("Rolled back.")
                    self.modificator.loadFromRemote()
                else:
                    print("Error sending data, server said: %s" % response["Message"])
        except Exception:
            _showTraceback()

    def do_mergeWithServer(self, dummy):
        """
        Shows diff with lastest version in server
        Usage: diffWithServer
        """
        try:
            choice = six.moves.input("Do you want to merge with server configuration? yes/no [no]: ")
            choice = choice.lower()
            if choice in ("yes", "y"):
                retVal = self.modificator.mergeWithServer()
                if retVal["OK"]:
                    print("Merged")
                else:
                    print("There was an error: ", retVal["Message"])
            else:
                print("Merge aborted")
        except Exception:
            _showTraceback()
